import logging
import os
import threading
import time
import warnings
from datetime import datetime
from multiprocessing import Lock
from pathlib import Path
from typing import Callable, Tuple, Union

import natsort as ns
import numpy as np
from matplotlib import pyplot as plt
from watchdog.events import (
    DirCreatedEvent,
    FileCreatedEvent,
    FileMovedEvent,
    FileSystemEventHandler,
)
from watchdog.observers import Observer

from toffy.json_utils import read_json_file


class RunStructure:
    """Expected bin and json files.

    Attributes:
        fov_progress (dict): Whether or not an expected file has been created
    """

    def __init__(self, run_folder: str, fov_timeout: int = 7800):
        """Initializes RunStructure by parsing run json within provided run folder.

        Args:
            run_folder (str):
                path to run folder
            fov_timeout (int):
                number of seconds to wait for non-null filesize before raising an error
        """
        self.timeout = fov_timeout
        self.fov_progress = {}
        self.processed_fovs = []
        self.moly_points = []

        # find run .json and get parameters
        run_name = Path(run_folder).parts[-1]
        run_metadata = read_json_file(
            os.path.join(run_folder, f"{run_name}.json"), encoding="utf-8"
        )

        # parse run_metadata and populate expected structure
        for fov in run_metadata.get("fovs", ()):
            run_order = fov.get("runOrder", -1)
            scan = fov.get("scanCount", -1)
            if run_order * scan < 0:
                raise KeyError(f"Could not locate keys in {run_folder}.json")

            # scan 2's don't contain significant imaging data per new MIBI specs
            fov_name = f"fov-{run_order}-scan-1"
            if fov.get("standardTarget", "") == "Molybdenum Foil":
                self.moly_points.append(fov_name)

            self.fov_progress[fov_name] = {"json": False, "bin": False}

        # get the highest FOV number, needed for checking if final FOV processed
        # NOTE: only scan-1 files considered, so len is good
        self.highest_fov = len(self.fov_progress)

    def check_run_condition(self, path: str) -> Tuple[bool, str]:
        """Checks if all requisite files exist and are complete.

        Args:
            path (str):
                path to expected file

        Raises:
            TimeoutError

        Returns:
            (bool, str):
                whether or not both json and bin files exist, as well as the name of the point
        """
        filename = Path(path).parts[-1]

        # if filename starts with a '.' (temp file), it should be ignored
        if filename[0] == ".":
            return False, ""

        # filename is not corrct format of fov.bin or fov.json
        if len(filename.split(".")) != 2:
            return False, ""

        fov_name, extension = filename.split(".")

        # path no longer valid
        if not os.path.exists(path):
            warnings.warn(
                f"{path} doesn't exist but was recently created. This should be unreachable...",
                Warning,
            )
            return False, ""

        # avoids repeated processing in case of duplicated events
        if fov_name in self.processed_fovs:
            return False, fov_name

        # does not process moly points, but add to process list to ensure proper incrementing
        if fov_name in self.moly_points:
            self.processed(fov_name)
            self.fov_progress[fov_name]["json"] = True
            self.fov_progress[fov_name]["bin"] = True
            return False, fov_name

        wait_time = 0
        if fov_name in self.fov_progress:
            if extension in self.fov_progress[fov_name]:
                while os.path.getsize(path) == 0:
                    # consider timed out fovs complete
                    if wait_time >= self.timeout:
                        del self.fov_progress[fov_name]
                        raise TimeoutError(f"timed out waiting for {path}...")

                    time.sleep(self.timeout / 10)
                    wait_time += self.timeout / 10

                self.fov_progress[fov_name][extension] = True

            if all(self.fov_progress[fov_name].values()):
                return True, fov_name

        elif extension == "bin":
            warnings.warn(f"Found unexpected bin file, {path}...", Warning)
            return False, ""

        return False, fov_name

    def processed(self, fov_name: str):
        """Notifies run structure that fov has been processed.

        Args:
            fov_name (str):
                Name of FoV
        """
        self.processed_fovs.append(fov_name)

    def check_fov_progress(self) -> dict:
        """Condenses internal dictionary to show which fovs have finished.

        Returns:
            dict
        """
        all_fovs = self.fov_progress.keys()
        moly_fovs = self.moly_points
        necessary_fovs = list(set(all_fovs).difference(moly_fovs))

        return {k: all(self.fov_progress[k].values()) for k in necessary_fovs}


class FOV_EventHandler(FileSystemEventHandler):
    """File event handler for FOV files.

    Attributes:
        run_folder (str):
            path to run folder
        watcher_out (str):
            folder to save all callback results + log file
        run_structure (RunStructure):
            expected run file structure + fov_progress status
        fov_callback (Callable[[str, str], None]):
            callback to run on each fov
        run_callback (Callable[[None], None]):
            callback to run over the entire run
    """

    instance_count = 0

    def __init__(
        self,
        run_folder: str,
        log_folder: str,
        fov_callback: Callable[[str, str], None],
        run_callback: Callable[[str], None],
        intermediate_callback: Callable[[str], None] = None,
        fov_timeout: int = 7800,
        watcher_timeout: int = 3 * 7800,
    ):
        """Initializes FOV_EventHandler.

        Args:
            run_folder (str):
                path to run folder
            log_folder (str):
                path to save outputs to
            fov_callback (Callable[[str, str], None]):
                callback to run on each fov
            run_callback (Callable[[None], None]):
                callback to run over the entire run
            intermediate_callback (Callable[[None], None]):
                run callback overriden to run on each fov
            fov_timeout (int):
                number of seconds to wait for non-null filesize before raising an error
            watcher_timeout (int):
                length to wait for new file generation before timing out
        """
        super().__init__()
        self.run_folder = run_folder

        self.last_event_time = datetime.now()
        self.timer_thread = threading.Thread(
            target=self.file_timer, args=(fov_timeout, watcher_timeout)
        )
        self.timer_thread.daemon = True
        self.timer_thread.start()

        self.log_path = os.path.join(log_folder, f"{Path(run_folder).parts[-1]}_log.txt")
        if not os.path.exists(log_folder):
            os.makedirs(log_folder)
        logging.basicConfig(
            level=logging.INFO,
            filename=self.log_path,
            filemode="a",
            format="%(name)s - %(levelname)s - %(message)s",
        )

        # create run structure
        self.run_structure = RunStructure(run_folder, fov_timeout=fov_timeout)

        self.fov_func = fov_callback
        self.run_func = run_callback
        self.inter_func = intermediate_callback
        self.inter_return_vals = None
        self.lock = Lock()
        self.last_fov_num_processed = 0
        self.all_fovs_complete = False

        for root, dirs, files in os.walk(run_folder):
            for name in ns.natsorted(files):
                # NOTE: don't call with check_last_fov to prevent duplicate processing
                self.on_created(FileCreatedEvent(os.path.join(root, name)), check_last_fov=False)

        # edge case if the last FOV gets written during the preprocessing stage
        # simulate a trigger using the first FOV file
        self._check_last_fov(
            os.path.join(root, list(self.run_structure.fov_progress.keys())[0] + ".bin")
        )

    def _check_fov_status(self, path: str):
        """Verifies the status of the file written at `path`.

        Args:
            path (str):
                The path to check the status of
        Returns:
            Tuple[Optional[str], Optional[str]]:
                The status of `path`, as well as the corresponding FOV name
        """
        try:
            fov_ready, point_name = self.run_structure.check_run_condition(path)
            return fov_ready, point_name
        except TimeoutError as timeout_error:
            print(f"Encountered TimeoutError error: {timeout_error}")
            logging.warning(
                f'{datetime.now().strftime("%d/%m/%Y %H:%M:%S")} -- '
                f"{path} never reached non-zero file size...\n"
            )

            # these count as processed FOVs, so increment
            self.last_fov_num_processed += 1

            # these count as processed FOVs, so mark as processed
            self.run_structure.processed(Path(path).parts[-1].split(".")[0])
            self.check_complete()

            return None, None

    def _generate_callback_data(self, point_name: str, overwrite: bool):
        """Runs the `fov_func` and `inter_func` if applicable for a FOV.

        Args:
            point_name (str):
                The name of the FOV to run FOV (and intermediate if applicable) callbacks on
            overwrite (bool):
                Forces an overwrite of already existing data, needed if a FOV needs re-extraction
        """
        print(f"Discovered {point_name}, beginning per-fov callbacks...")
        logging.info(f'{datetime.now().strftime("%d/%m/%Y %H:%M:%S")} -- Extracting {point_name}\n')
        logging.info(
            f'{datetime.now().strftime("%d/%m/%Y %H:%M:%S")} -- '
            f"Running {self.fov_func.__name__} on {point_name}\n"
        )

        self.fov_func(self.run_folder, point_name, overwrite)
        self.run_structure.processed(point_name)

        if self.inter_func:
            # clear plots contained in intermediate return values if set
            if self.inter_return_vals:
                qc_plots = self.inter_return_vals.get("plot_qc_metrics", None)
                mph_plot = self.inter_return_vals.get("plot_mph_metrics", None)

                if qc_plots or mph_plot:
                    plt.cla()
                    plt.clf()
                    plt.close("all")

            self.inter_return_vals = self.inter_func(self.run_folder)

        self.check_complete()

    def _process_missed_fovs(self, path: str):
        """Given a `path`, check if there are any missing FOVs to process before it.

        Args:
            path (str):
                The path to check for missing FOVs prior
        """
        # verify the path provided is correct .bin type, if not skip
        filename = Path(path).parts[-1]
        name_ext = filename.split(".")
        if len(name_ext) != 2 or name_ext[1] != "bin":
            return

        # NOTE: MIBI now only stores relevant data in scan 1, ignore any scans > 1
        scan_num = int(name_ext[0].split("-")[3])
        if scan_num > 1:
            return

        # retrieve the FOV number
        fov_num = int(name_ext[0].split("-")[1])

        # a difference of 1 from last_fov_num_processed means there are no in-between FOVs
        # set to <= for safety (this should never happen in theory)
        if fov_num - 1 <= self.last_fov_num_processed:
            return

        # NOTE: from observation, only the most recent FOV will ever be in danger of timing out
        # so all the FOVs processed in this function should already be fully processed
        start_index = self.last_fov_num_processed + 1 if self.last_fov_num_processed else 1
        for i in np.arange(start_index, fov_num):
            fov_name = f"fov-{i}-scan-1"
            fov_bin_file = os.path.join(self.run_folder, fov_name + ".bin")
            fov_json_file = os.path.join(self.run_folder, fov_name + ".json")

            # this can happen if there's a lag copying files over
            while not os.path.exists(fov_bin_file) and not os.path.exists(fov_json_file):
                time.sleep(60)

            self._fov_callback_driver(os.path.join(self.run_folder, fov_name + ".bin"))
            self._fov_callback_driver(os.path.join(self.run_folder, fov_name + ".json"))

    def _check_last_fov(self, path: str):
        """Checks if the last FOV's data has been written.

        Needed because there won't be any more file triggers after this happens.

        Args:
            path (str):
                The path that triggers this call. Used only for formatting purposes.
        """
        # define the name of the last FOV
        last_fov = f"fov-{self.run_structure.highest_fov}-scan-1"
        last_fov_bin = f"{last_fov}.bin"
        last_fov_json = f"{last_fov}.json"

        # if the last FOV has been written, then process everything up to that if necessary
        # NOTE: don't process if it has already been written
        bin_dir = str(Path(path).parents[0])
        last_fov_is_processed = self.last_fov_num_processed == self.run_structure.highest_fov
        last_fov_data_exists = os.path.exists(
            os.path.join(bin_dir, last_fov_bin)
        ) and os.path.exists(os.path.join(bin_dir, last_fov_json))

        if not last_fov_is_processed and last_fov_data_exists:
            start_index = self.last_fov_num_processed + 1 if self.last_fov_num_processed else 1
            for i in np.arange(start_index, self.run_structure.highest_fov):
                fov_name = f"fov-{i}-scan-1"
                fov_bin_file = os.path.join(self.run_folder, fov_name + ".bin")
                fov_json_file = os.path.join(self.run_folder, fov_name + ".json")

                # this can happen if there's a lag copying files over
                while not os.path.exists(fov_bin_file) and not os.path.exists(fov_json_file):
                    time.sleep(60)

                self._fov_callback_driver(os.path.join(self.run_folder, fov_name + ".bin"))
                self._fov_callback_driver(os.path.join(self.run_folder, fov_name + ".json"))

            # process the final bin file
            self._fov_callback_driver(os.path.join(self.run_folder, last_fov_bin))
            self._fov_callback_driver(os.path.join(self.run_folder, last_fov_json))

            # explicitly call check_complete to start run callbacks, since all FOVs are done
            self.check_complete()

    def _check_bin_updates(self):
        """Checks for, and re-runs if necessary, any incompletely extracted FOVs."""
        for fov in self.run_structure.fov_progress:
            # skip moly points
            if fov in self.run_structure.moly_points:
                continue

            fov_bin_path = os.path.join(self.run_folder, fov + ".bin")
            fov_json_path = os.path.join(self.run_folder, fov + ".json")

            # if .bin file ctime > .json file ctime, incomplete extraction, need to re-extract
            fov_bin_create = Path(fov_bin_path).stat().st_ctime
            fov_json_create = Path(fov_json_path).stat().st_ctime

            if fov_bin_create > fov_json_create:
                warnings.warn(f"Re-extracting incompletely extracted FOV {fov}")
                logging.info(
                    f'{datetime.now().strftime("%d/%m/%Y %H:%M:%S")} -- Re-extracting {fov}\n'
                )
                logging.info(
                    f'{datetime.now().strftime("%d/%m/%Y %H:%M:%S")} -- '
                    f"Running {self.fov_func.__name__} on {fov}\n"
                )

                # since reprocessing needed, remove from self.processed_fovs
                self.run_structure.processed_fovs.remove(fov)

                # re-extract the .bin file
                # NOTE: since no more FOVs are being written, last_fov_num_processed is irrelevant
                self._fov_callback_driver(fov_bin_path, overwrite=True)

    def _fov_callback_driver(self, file_trigger: str, overwrite: bool = False):
        """The FOV and intermediate-level callback motherbase for a single .bin file.

        Args:
            file_trigger (str):
                The file that gets caught by the watcher to throw into the pipeline
            overwrite (bool):
                Forces an overwrite of already existing data, needed if a FOV needs re-extraction
        """
        # check if what's created is in the run structure
        fov_ready, point_name = self._check_fov_status(file_trigger)

        if fov_ready:
            self._generate_callback_data(point_name, overwrite=overwrite)

        # needs to update if .bin file processed OR new moly point detected
        is_moly = point_name in self.run_structure.moly_points
        is_processed = point_name in self.run_structure.processed_fovs
        if fov_ready or (is_moly and not is_processed):
            self.last_fov_num_processed += 1

    def _run_callbacks(
        self, event: Union[DirCreatedEvent, FileCreatedEvent, FileMovedEvent], check_last_fov: bool
    ):
        """The pipeline runner, invoked when a new event is seen.

        Args:
            event (Union[DirCreatedEvent, FileCreatedEvent, FileMovedEvent]):
                The type of event seen. File/directory creation and file renaming are supported.
            check_last_fov (bool):
                Whether to invoke `_check_last_fov` on the event
        """
        if type(event) in [DirCreatedEvent, FileCreatedEvent]:
            file_trigger = event.src_path
        else:
            file_trigger = event.dest_path

        # process any FOVs that got missed on the previous iteration of on_created/on_moved
        self._process_missed_fovs(file_trigger)

        # run the fov callback process on the file
        self._fov_callback_driver(file_trigger)

        if check_last_fov:
            self._check_last_fov(file_trigger)

    def on_created(self, event: FileCreatedEvent, check_last_fov: bool = True):
        """Handles file creation events.

        If FOV structure is completed, the fov callback, `self.fov_func` will be run over the data.
        This function is automatically called; users generally shouldn't call this function

        Args:
            event (FileCreatedEvent): file creation event
            check_last_fov (bool): whether to check if the last fov has been processed
        """
        # reset event creation time
        current_time = datetime.now()
        self.last_event_time = current_time

        # this happens if _check_last_fov gets called by a prior FOV, no need to reprocess
        if self.last_fov_num_processed == self.run_structure.highest_fov:
            return

        with self.lock:
            super().on_created(event)
            self._run_callbacks(event, check_last_fov)

    def file_timer(self, fov_timeout, watcher_timeout):
        """Checks time since last file was generated.

        Args:
            fov_timeout (int):
                how long to wait for fov data to be generated once file detected
            watcher_timeout (int):
                length to wait for new file generation before timing out
        """
        while True:
            current_time = datetime.now()
            time_elapsed = (current_time - self.last_event_time).total_seconds()

            # 3 fov cycles and no new files --> timeout
            if time_elapsed > watcher_timeout:
                fov_num = self.last_fov_num_processed
                fov_name = list(self.run_structure.fov_progress.keys())[fov_num]
                print(f"Timed out waiting for {fov_name} files to be generated.")
                logging.info(
                    f'{datetime.now().strftime("%d/%m/%Y %H:%M:%S")} -- Timed out'
                    f"waiting for {fov_name} files to be generated.\n"
                )
                logging.info(
                    f'{datetime.now().strftime("%d/%m/%Y %H:%M:%S")} -- '
                    f"Running {self.run_func.__name__} on FOVs\n"
                )

                # mark remaining fovs as completed to exit watcher
                for fov_name in list(self.run_structure.fov_progress.keys()):
                    self.run_structure.fov_progress[fov_name] = {"json": True, "bin": True}

                # trigger run callbacks
                self.run_func(self.run_folder)
                break
            time.sleep(fov_timeout)

    def on_moved(self, event: FileMovedEvent, check_last_fov: bool = True):
        """Handles file renaming events.

        If FOV structure is completed, the fov callback, `self.fov_func` will be run over the data.
        This function is automatically called; users generally shouldn't call this function

        Args:
            event (FileMovedEvent): file moved event
            check_last_fov (bool): whether to check if last fov was processed
        """
        # this happens if _check_last_fov gets called by a prior FOV, no need to reprocess
        if self.last_fov_num_processed == self.run_structure.highest_fov:
            return

        with self.lock:
            super().on_moved(event)
            self._run_callbacks(event, check_last_fov)

    def check_complete(self):
        """Checks run structure fov_progress status.

        If run is complete, all callbacks in `per_run` will be run over the whole run.

        NOTE: bin files that had new data written will first need to be re-extracted.
        """
        if all(self.run_structure.check_fov_progress().values()) and not self.all_fovs_complete:
            self.all_fovs_complete = True
            self._check_bin_updates()
            logging.info(f'{datetime.now().strftime("%d/%m/%Y %H:%M:%S")} -- All FOVs finished\n')
            logging.info(
                f'{datetime.now().strftime("%d/%m/%Y %H:%M:%S")} -- '
                f"Running {self.run_func.__name__} on whole run\n"
            )

            self.run_func(self.run_folder)


def start_watcher(
    run_folder: str,
    log_folder: str,
    fov_callback: Callable[[str, str], None],
    run_callback: Callable[[None], None],
    intermediate_callback: Callable[[str, str], None] = None,
    run_folder_timeout: int = 5400,
    completion_check_time: int = 30,
    zero_size_timeout: int = 7800,
    watcher_timeout: int = 3 * 7800,
):
    """Passes bin files to provided callback functions as they're created.

    Args:
        run_folder (str):
            path to run folder
        log_folder (str):
            where to create log file
        fov_callback (Callable[[str, str], None]):
            function to run on each completed fov. assemble this using
            `watcher_callbacks.build_callbacks`
        run_callback (Callable[[None], None]):
            function ran once the run has completed. assemble this using
            `watcher_callbacks.build_callbacks`
        intermediate_callback (Callable[[None], None]):
            function defined as run callback overriden as fov callback. assemble this using
            `watcher_callbacks.build_callbacks`
        run_folder_timeout (int):
            how long to wait for the run folder to appear before timing out, in seconds.
            note that the watcher cannot begin until this run folder appears.
        completion_check_time (int):
            how long to wait before checking watcher completion, in seconds.
            note, this doesn't effect the watcher itself, just when this wrapper function exits.
        zero_size_timeout (int):
            number of seconds to wait for non-zero file size
        watcher_timeout (int):
            number of seconds to wait for new fov file generation
    """
    # if the run folder specified isn't already there, ask the user to explicitly confirm the name
    if not os.path.exists(run_folder):
        warnings.warn(
            f"Waiting for {run_folder}. Please first double check that your run data "
            "doesn't already exist under a slightly different name in D:\\Data. "
            "Sometimes, the CACs change capitalization or add extra characters to the run folder. "
            "If this happens, stop the watcher and update the run_name variable in the notebook "
            "before trying again."
        )

    # allow the watcher to poll the run folder until it appears or times out
    run_folder_wait_time = 0
    while not os.path.exists(run_folder) and run_folder_wait_time < run_folder_timeout:
        time.sleep(run_folder_timeout / 10)
        run_folder_wait_time += run_folder_timeout / 10

    if run_folder_wait_time == run_folder_timeout:
        raise FileNotFoundError(
            f"Timed out waiting for {run_folder}. Make sure the run_name variable in the notebook "
            "matches up with the run folder name in D:\\Data, or try again a few minutes later "
            "if the run folder still hasn't shown up."
        )

    observer = Observer()
    event_handler = FOV_EventHandler(
        run_folder,
        log_folder,
        fov_callback,
        run_callback,
        intermediate_callback,
        zero_size_timeout,
        watcher_timeout,
    )
    observer.schedule(event_handler, run_folder, recursive=True)
    observer.start()

    try:
        while not all(event_handler.run_structure.check_fov_progress().values()):
            time.sleep(completion_check_time)
    except KeyboardInterrupt:
        observer.stop()

    observer.stop()
    observer.join()



import os

import numpy as np
import pandas as pd
from alpineer import io_utils

necessary_masses = pd.DataFrame(
    {
        "Mass": [
            39,
            48,
            56,
            69,
            71,
            89,
            113,
            115,
            117,
            141,
            142,
            143,
            144,
            145,
            146,
            147,
            148,
            149,
            150,
            151,
            152,
            153,
            154,
            155,
            156,
            157,
            158,
            159,
            160,
            161,
            162,
            163,
            164,
            165,
            166,
            167,
            168,
            169,
            170,
            171,
            172,
            173,
            174,
            175,
            176,
            197,
        ],
        "Target": [
            "chan_39",
            "chan_48",
            "Fe",
            "chan_69",
            "chan_71",
            "chan_89",
            "chan_113",
            "chan_115",
            "Noodle",
            "chan_141",
            "chan_142",
            "chan_143",
            "chan_144",
            "chan_145",
            "chan_146",
            "chan_147",
            "chan_148",
            "chan_149",
            "chan_150",
            "chan_151",
            "chan_152",
            "chan_153",
            "chan_154",
            "chan_155",
            "chan_156",
            "chan_157",
            "chan_158",
            "chan_159",
            "chan_160",
            "chan_161",
            "chan_162",
            "chan_163",
            "chan_164",
            "chan_165",
            "chan_166",
            "chan_167",
            "chan_168",
            "chan_169",
            "chan_170",
            "chan_171",
            "chan_172",
            "chan_173",
            "chan_174",
            "chan_175",
            "chan_176",
            "Au",
        ],
    }
)


def modify_panel_ranges(panel: pd.DataFrame, start_offset: float = 0, stop_offset: float = 0):
    """Adjust the offsets of a given panel.

    Only applicable for masses with ranges separated by 0.3 between 'Stop' and 'Start'.

    Args:
        panel (pd.DataFrame): panel dataframe with columns Mass, Target, Start, and Stop.
        start_offset (float): the value to add to the `'Start'` column.
        stop_offset (float): the value to add to the `'Stop'` column.

    Returns:
        pd.DataFrame:
            Updated panel with `start_offset` added to `'Start`' column,
            likewise for `stop_offset` and `'Stop'` column.
    """
    panel_new = panel.copy()

    # extract only rows where 'Start' - 'End' = -0.3, round to account for floating point error
    panel_rows_modify = panel_new[
        (panel_new["Start"] - panel_new["Stop"]).round(1) == -0.3
    ].index.values

    # ensure safe conversion of types
    panel_new["Start"] = panel_new["Start"].astype(np.float64)
    panel_new["Stop"] = panel_new["Stop"].astype(np.float64)

    # add start_offset to 'Start' column
    panel_new.loc[panel_rows_modify, "Start"] = (
        panel_new.loc[panel_rows_modify, "Start"].copy() + start_offset
    )

    # add stop_offset to 'Stop' column
    panel_new.loc[panel_rows_modify, "Stop"] = (
        panel_new.loc[panel_rows_modify, "Stop"].copy() + stop_offset
    )

    return panel_new


def merge_duplicate_masses(panel):
    """Check a panel df for duplicate mass values and return a unique mass panel with the
        target names combined.

    Args:
        panel (pd.DataFrame): panel dataframe with columns Mass and Target
    Returns:
        pd.DataFrame with no duplicate masses
    """
    # find the mass and target values of duplicate
    duplicates = panel[panel["Mass"].duplicated(keep=False)]

    # combine target names
    for mass in duplicates["Mass"].unique():
        duplicate_targets = panel.loc[panel["Mass"] == mass, "Target"]
        target_name = "".join(str(name) + "_" for name in duplicate_targets)[:-1]
        panel.loc[panel["Mass"] == mass, "Target"] = target_name

    # only keep one row for each mass value
    unique_panel = panel.drop_duplicates(subset=["Mass"], keep="first")

    return unique_panel


def convert_panel(panel_path):
    """Converts the panel retrieved from ionpath into a necessary toffy format,
    also adds necessary channels for analysis.

    Args:
        panel_path (str): direct path to panel file
    Returns:
        panel (pd.DataFrame): detailing the Mass, Target, Start, and Stop values
        also saves the above to file with name formatted "panel_name-toffy.csv"
    """
    # retrieve the panel name and directory it is contained in
    panel_name = os.path.basename(os.path.splitext(panel_path)[0])
    panel_dir = os.path.dirname(panel_path)

    # read in metadata contained in the first few lines
    col_index = 0
    with open(panel_path, encoding="utf-8") as r:
        lines = r.readlines()
        for index, line in enumerate(lines):
            # line containing the column names we want
            if len(line.split(",")) == 10:
                col_index = index
                break
        r.close()

    # load in the panel info to df
    with open(panel_path, encoding="utf-8-sig") as r:
        headers = [next(r) for i in list(range(col_index + 1))]
        cols = headers[col_index].split(",")
        panel = pd.read_csv(r, sep=",", names=cols, index_col=False)
        r.close()

    panel.columns = [panel.columns.str.replace('"', "")]
    panel.columns = panel.columns.get_level_values(0)

    # check for already correctly formatted panel, return as is
    if list(panel.columns) == ["Mass", "Target", "Start", "Stop\n"]:
        print(f"{panel_name}.csv has the correct toffy format. Loading in panel data.")
        panel.columns = ["Mass", "Target", "Start", "Stop"]
        return panel
    # if not ionpath panel, raise error
    elif list(panel.columns) != [
        "ID (Lot)",
        "Target",
        "Clone",
        "Mass",
        "Element",
        "Manufactured",
        "Stock",
        "Titer",
        "Volume (μL)",
        "Staining Batch\n",
    ]:
        raise ValueError(f"{panel_name}.csv is not an Ionpath or toffy structured panel.")

    # retrieve original mass / target values
    toffy_panel = panel[["Mass", "Target"]].copy()

    # include only the unique provided masses with concatenated target names
    toffy_panel = merge_duplicate_masses(toffy_panel)

    # add necessary panel masses to the toffy panel, removing any duplicates
    toffy_panel = pd.concat([toffy_panel, necessary_masses], ignore_index=True)
    toffy_panel = toffy_panel.drop_duplicates(subset=["Mass"], keep="first")

    # edit panel to add Start and Stop values
    toffy_panel["Start"] = toffy_panel["Mass"].copy() - 0.3
    toffy_panel["Stop"] = toffy_panel["Mass"].copy()

    # specify 116.7 - 125 Noodle extraction range
    toffy_panel.loc[toffy_panel.Mass == 117, "Stop"] = 125

    # sort data by mass
    toffy_panel = toffy_panel.sort_values(by=["Mass"])

    toffy_panel.to_csv(os.path.join(panel_dir, panel_name + "-toffy.csv"), index=False)
    print(
        f"{panel_name}.csv does not have the correct toffy format. "
        f"Creating {panel_name}-toffy.csv and loading in panel data."
    )

    return toffy_panel


def load_panel(panel_path):
    """Loads in the toffy panel data, calls convert_panel() if necessary.

    Args:
        panel_path (str): direct path to panel file
    Returns:
        panel (pd.DataFrame): toffy formatted panel data
    """
    # read in the provided panel info
    panel_name = os.path.basename(panel_path).split(".")[0]
    panel_dir = os.path.dirname(panel_path)

    # if panel path points to toffy panel, load it
    if "-toffy" in panel_name:
        toffy_panel = pd.read_csv(panel_path, index_col=False)
        if list(toffy_panel.columns) != ["Mass", "Target", "Start", "Stop"]:
            raise ValueError(
                f"{panel_name}.csv is not correctly formatted. Please remove "
                "'-toffy' from the file name."
            )
        else:
            toffy_panel.loc[toffy_panel.Mass == 117, "Stop"] = 125
            print(f"{panel_name}.csv is in the correct toffy format. Loading in panel data.")
    else:
        # check if toffy panel exists in panel_dir and load it
        files = io_utils.list_files(panel_dir, substrs=panel_name)
        if panel_name + "-toffy.csv" in files:
            converted_panel_path = os.path.join(panel_dir, panel_name + "-toffy.csv")
            toffy_panel = pd.read_csv(converted_panel_path, index_col=False)
            toffy_panel.loc[toffy_panel.Mass == 117, "Stop"] = 125
            print(
                f"Detected {panel_name}-toffy.csv in {panel_dir}. "
                "Loading in toffy formatted panel data."
            )
        # if no toffy panel, create one
        else:
            toffy_panel = convert_panel(panel_path)

    return toffy_panel



import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from alpineer import io_utils
from mibi_bin_tools import bin_files
from natsort import natsort_keygen

from toffy.json_utils import read_json_file
from toffy.normalize import combine_run_metrics


def get_estimated_time(bin_file_dir, fov):
    """Retrieve run time data for each fov json file.

    Args:
        bin_file_dir (str): path to the FOV bin and json files
        fov (str): name of fov to get estimated time for
    Returns:
        fov_time (int): estimated run time for the given fov
    """
    # path validation
    io_utils.validate_paths(bin_file_dir)

    # get fov json file in bin_file_path
    json_file = os.path.join(bin_file_dir, f"{fov}.json")
    if not os.path.exists(json_file):
        raise FileNotFoundError(f"The FOV name supplied doesn't have a JSON file: {fov}")

    # retrieve estimated time (frame dimensions x pixel dwell time)
    run_metadata = read_json_file(json_file, encoding="utf-8")
    try:
        size = run_metadata.get("frameSize")
        time = run_metadata.get("dwellTimeMillis")
        estimated_time = int(size**2 * time)
    except TypeError:
        raise KeyError(
            "The FOV json file is missing one of the necessary keys (frameSize or dwellTimeMillis)"
        )

    return estimated_time


def generate_time_ticks(mph_df):
    """Create a time axis for median pulse heights with ticks at approx. 6 hour increments.

    Args:
         mph_df: contains mph date, specifically requires cum_total_count and cum_total_time
         columns
    Returns:
        list of two lists detailing tick locations and tick number labels
    """
    # determine number of ticks and what the labels should be based on total run time
    sub_df = mph_df[["cum_total_count", "cum_total_time"]]
    total_time = sub_df["cum_total_time"].iloc[-1]
    tick_num = int(total_time / (6 * (3600 * 1000)))
    tick_labels = [i * 6 for i in range(0, tick_num + 1)]
    time_ticks = [tick * (3600 * 1000) for tick in tick_labels[1 : len(tick_labels)]]

    # find count value associated with the time closest to each tick
    tick_locations = [0]
    for tick in time_ticks:
        count_tick = (sub_df.iloc[(sub_df["cum_total_time"] - tick).abs().argsort()[:1]])[
            "cum_total_count"
        ]
        count_tick = (count_tick.to_string()).split(" ")[4]
        tick_locations.append(int(count_tick) / 1000000)

    return [tick_locations, tick_labels]


def compute_mph_metrics(bin_file_dir, csv_dir, fov, mass=98, mass_start=97.5, mass_stop=98.5):
    """Retrieves total counts, pulse heights, & estimated time for a given FOV.

    Args:
        bin_file_dir (str): path to the FOV bin and json files
        csv_dir (str): path to output csv to
        fov (string): name of fov bin file without the extension
        mass (float): mass for the panel
        mass_start (float): beginning of mass integration range
        mass_stop (float): end of mass integration range
    """
    target = None
    panel = pd.DataFrame(
        [
            {
                "Mass": mass,
                "Target": target,
                "Start": mass_start,
                "Stop": mass_stop,
            }
        ]
    )

    # retrieve the data from bin file and output to individual csv
    pulse_height_file = fov + "-mph_pulse.csv"

    try:
        median = list(
            bin_files.get_median_pulse_height(bin_file_dir, fov, [target], panel).values()
        )[0]
        count_dict = bin_files.get_total_counts(bin_file_dir, [fov])
    except FileNotFoundError:
        raise FileNotFoundError(f"The FOV name supplied doesn't have a JSON file: {fov}")

    count = count_dict[fov]
    time = get_estimated_time(bin_file_dir, fov)

    out_df = pd.DataFrame({"fov": [fov], "MPH": [median], "total_count": [count], "time": [time]})

    # saves individual .csv files to csv_dir
    out_df.to_csv(os.path.join(csv_dir, pulse_height_file), index=False)


def combine_mph_metrics(csv_dir, return_data=False, warn_overwrite=True):
    """Combines data from individual csvs into one.

    Args:
        csv_dir (str): path where FOV mph data csvs are stored
        return_data (bool): whether to return dataframe with mph metrics, default False
        warn_overwrite (bool): whether to warn if existing `_combined.csv` file found

    Returns:
        combined mph data for all FOVs
    """
    # path validation checks
    io_utils.validate_paths(csv_dir)

    # combine individual csv files
    combine_run_metrics(csv_dir, "mph_pulse", warn_overwrite)

    # calculate cumulative sums of total counts and time
    combined_df = pd.read_csv(os.path.join(csv_dir, "mph_pulse_combined.csv"))
    combined_df = combined_df.sort_values(by="fov", key=natsort_keygen())
    combined_df["cum_total_count"] = combined_df["total_count"].cumsum()
    combined_df["cum_total_time"] = combined_df["time"].cumsum()

    combined_df.to_csv(os.path.join(csv_dir, "mph_pulse_combined.csv"), index=False)

    # return data
    if return_data:
        return combined_df


def visualize_mph(mph_df, out_dir, regression: bool = False, return_plot: bool = False):
    """Create a scatterplot visualizing median pulse heights by FOV cumulative count
    Args:
        mph_df (pd.DataFrame): data detailing total counts and pulse heights
        out_dir (str): path of directory to save plot to
        regression (bool): whether to plot regression line, default is False
        return_plot (bool): if True, this will return the plot. Defaults to False.
    """
    # path validation checks
    if out_dir is not None:
        io_utils.validate_paths(out_dir)

    # visualize the median pulse heights
    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    x = mph_df["cum_total_count"] / 1000000
    y = mph_df["MPH"]
    ax1.set_xlabel("FOV cumulative count (in millions)")
    ax1.set_ylabel("median pulse height")
    ax1.scatter(x, y)
    ax2 = ax1.twiny()
    ax2.set_xlabel("estimated time (hours)")
    plt.title("FOV total counts vs median pulse height")

    # create time axis
    new_ticks = generate_time_ticks(mph_df)
    tick_locations = new_ticks[0]
    tick_labels = new_ticks[1]
    ax2.set_xlim(ax1.get_xlim())
    ax2.set_xticks(tick_locations)
    ax2.set_xticklabels(tick_labels)
    plt.gcf().set_size_inches(18.5, 10.5)

    # plot regression line
    if regression:
        # plot with regression line
        x2 = np.array(mph_df["cum_total_count"] / 1000000)
        y2 = np.array(mph_df["MPH"])
        m, b = np.polyfit(x2, y2, 1)
        ax1.plot(x2, m * x2 + b)

    # save figure
    file_path = os.path.join(out_dir, "fov_vs_mph.png")
    if os.path.exists(file_path):
        os.remove(file_path)
    plt.savefig(file_path)

    if return_plot:
        return fig



import os
import re
import warnings

import natsort as ns
import numpy as np
import pandas as pd
from alpineer import io_utils, load_utils
from mibi_bin_tools import bin_files

from toffy.json_utils import check_for_empty_files, list_moly_fovs, read_json_file


def extract_missing_fovs(
    bin_file_dir,
    extraction_dir,
    panel,
    extract_intensities=["Au", "chan_39"],
    replace=True,
):
    """Check for already extracted FOV bin files, and extract the remaining (excluding moly fovs
       and fovs with empty json files).

    Args:
        bin_file_dir (str): path to directory containing the bin and json files
        extraction_dir (str): path to directory of already extracted FOVs
        panel (pd.DataFrame): file defining the panel info for bin file extraction
        extract_intensities (bool): whether to extract intensities from the bin files
        replace (bool): whether to replace pulse images with intensity
    """
    # retrieve all fov names from base_dir and extracted fovs from extraction_dir
    fovs = io_utils.remove_file_extensions(io_utils.list_files(bin_file_dir, substrs=".bin"))
    extracted_fovs = io_utils.list_folders(extraction_dir, substrs="fov")

    # filter out empty json file fovs
    empty_fovs = check_for_empty_files(bin_file_dir)
    if empty_fovs:
        fovs = list(set(fovs).difference(empty_fovs))

    if len(fovs) == 0:
        warnings.warn(f"No viable bin files were found in {bin_file_dir}", Warning)
        return

    # check for moly fovs
    moly_fovs = list_moly_fovs(bin_file_dir, fovs)

    if extracted_fovs:
        print(
            "Skipping the following previously extracted FOVs: ",
            ", ".join(extracted_fovs),
        )
    if moly_fovs:
        print("Moly FOVs which will not be extracted: ", ", ".join(moly_fovs))
    if empty_fovs:
        print(
            "FOVs with empty json files which will not be extracted: ",
            ", ".join(empty_fovs),
        )

    # extract missing fovs to extraction_dir
    non_moly_fovs = list(set(fovs).difference(moly_fovs))
    missing_fovs = list(set(non_moly_fovs).difference(extracted_fovs))
    missing_fovs = ns.natsorted(missing_fovs)

    if missing_fovs:
        print(f"Found {len(missing_fovs)} FOVs to extract.")
        bin_files.extract_bin_files(
            bin_file_dir,
            extraction_dir,
            include_fovs=missing_fovs,
            panel=panel,
            intensities=extract_intensities,
            replace=replace,
        )
        print("Extraction completed!")
    else:
        warnings.warn(f"No viable bin files were found in {bin_file_dir}", Warning)


def incomplete_fov_check(
    bin_file_dir, extraction_dir, num_rows=10, num_channels=5, signal_percent=0.02
):
    """Read in the supplied number tiff files for each FOV to check for incomplete images.

    Args:
        bin_file_dir (str): directory containing the run json file
        extraction_dir (str): directory containing the extracted tifs
        num_rows (int): number of bottom rows of the images to check for zero values
        num_channels (int): number of channel images to check per FOV
        signal_percent (float): min amount of non-zero signal required for complete FOVs

    Raises:
        Warning if any FOVs have only partially generated images
    """
    io_utils.validate_paths([bin_file_dir, extraction_dir])

    # read in json file to get custom fov names
    run_name = os.path.basename(bin_file_dir)
    run_file_path = os.path.join(bin_file_dir, run_name + ".json")
    run_metadata = read_json_file(run_file_path, encoding="utf-8")

    # get fov and channel info
    fovs = io_utils.list_folders(extraction_dir, "fov")
    channels = io_utils.list_files(os.path.join(extraction_dir, fovs[0]), ".tiff")
    channels_subset = channels[:num_channels]
    if "Au.tiff" not in channels_subset:
        channels_subset = channels_subset[:-1] + ["Au.tiff"]

    incomplete_fovs = {}
    for fov in fovs:
        # load in channel images
        img_data = load_utils.load_imgs_from_tree(
            extraction_dir, fovs=[fov], channels=channels_subset
        )
        row_index = img_data.shape[1] - num_rows
        img_bottoms = img_data[0, row_index:, :, :]

        # check percentage of non-zero pixels in the bottom of the image
        total_pixels = img_data.shape[1] * num_rows * num_channels
        if np.count_nonzero(img_bottoms) / total_pixels < signal_percent:
            i = re.findall(r"\d+", fov)[0]
            custom_name = run_metadata["fovs"][int(i) - 1]["name"]
            incomplete_fovs[fov] = custom_name

    if incomplete_fovs:
        incomplete_fovs = pd.DataFrame(incomplete_fovs, index=[0]).T
        incomplete_fovs.columns = ["fov_name"]
        warnings.warn(
            f"\nThe following FOVs have less than {signal_percent*100}% positive signal on average"
            " at the bottom of the tiff and may have been only partially imaged: \n"
            f"{incomplete_fovs}\n"
        )



import os
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Tuple

import numpy as np
import pandas as pd
import xarray as xr
from alpineer.image_utils import save_image
from skimage import draw, exposure, filters, measure, restoration, util


@dataclass
class StreakData:
    """Contains data for correcting the streaks consisting of binary masks, dataframes with
    location and size properties, a directory for saving, and the shape / channel for mask
    generation.

    Args:
        shape (tuple): The shape of the image / fov.
        fov (str): The name of the fov being processed.
        streak_channel (str): The specific channel name used to create the masks.
        corrected_dir (Path): The directory used to save the corrected tiffs and data in.
        streak_mask (np.ndarray): The first binary mask indicating candidate streaks.
        streak_df (pd.DataFrame): A dataframe, containing the location, area, and eccentricity
        of each streak.
        filtered_streak_mask (np.ndarray): A binary mask with out the false streaks.
        filtered_streak_df (pd.DataFrame): A subset of the `streak_df` containing location, area
        and eccentricity values of the filtered streaks.
        boxed_streaks (np.ndarray): An optional binary mask containing an outline for each
        filtered streaks.
        corrected_streak_mask (np.ndarray): An optional binary mask containing the lines used for
        correcting the streaks.
    """

    shape: Tuple[int, int] = None
    fov: str = None
    streak_channel: str = None
    corrected_dir: Path = None
    streak_mask: np.ndarray = None
    streak_df: pd.DataFrame = None
    filtered_streak_mask: np.ndarray = None
    filtered_streak_df: pd.DataFrame = None
    boxed_streaks: np.ndarray = None
    corrected_streak_mask: np.ndarray = None


def _get_save_dir(data_dir: Path, name: str, ext: str) -> Path:
    """A helper function which generates the path where the masks and DataFrames
    are saved to.

    Args:
        data_dir (Path): The directory to save the binary masks and DataFrames.
        name (str): The field of the DataClass to be saved.
        ext (str): The file extension, either `csv` or `tiff`.

    Returns:
        Path: Returns the path where the file is saved to.
    """
    return Path(data_dir, name + f".{ext}")


def _save_streak_data(streak_data: StreakData, name: str):
    """Helper function for saving tiff binary masks and dataframes.

    Args:
        streak_data (StreakData): An instance of the StreakData Dataclass, holds all necessary
        data for streak correction.
        name (str): The field of the DataClass to be saved.
    """
    data_dir = streak_data.corrected_dir / f"streak_data_{streak_data.streak_channel}"
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    data = getattr(streak_data, name)
    st = partial(_get_save_dir, data_dir, name)

    if type(data) is np.ndarray:
        save_image(st("tiff"), data)
    elif type(data) is pd.DataFrame:
        data.to_csv(st("csv"), index=True)


def _save_streak_masks(streak_data: StreakData):
    """Saves the data in StreakData as a tiff file if it's a Numpy array, and a csv if it is a
    Pandas DataFrame. Useful for visualization and debugging.

    Args:
        streak_data (StreakData): An instance of the StreakData Dataclass, holds all necessary
        data for streak correction.
    """
    fields = [
        "streak_mask",
        "streak_df",
        "filtered_streak_mask",
        "filtered_streak_df",
        "boxed_streaks",
        "corrected_streak_mask",
    ]
    for field in fields:
        _save_streak_data(streak_data, name=field)


def _make_binary_mask(
    input_image: np.ndarray,
    gaussian_sigma: float = 5.00,
    gamma: float = 4.0,
    gamma_gain: float = 1.00,
    log_gain: float = 1.00,
    pmin: int = 2,
    pmax: int = 98,
    threshold: float = 0.35,
    wavelet: str = "db2",
    mode: str = "soft",
    rescale_sigma: bool = True,
) -> np.ndarray:
    """Performs a series of denoiseing, filtering, and exposure adjustments to create a binary
    mask for the given input image.

    Args:
        input_image (np.ndarray): The image to perform the streak masking on.
        gaussian_sigma (float, optional): Parameter for `skimage.filters.gaussian`. Defaults to
        5.00.
        gamma (float, optional): Parameter for `skimage.exposure.adjust_gamma`. Defaults to 3.80.
        gamma_gain (float, optional): Parameter for `skimage.exposure.adjust_gamma`. Defaults to
        0.10.
        log_gain (float, optional): Parameter for `skimage.exposure.adjust_log`. Defaults to 1.00.
        pmin (int, optional): Lower bound for the `np.percentile` threshold, used for rescaling
        the intensity. Defaults to 2.
        pmax (int, optional): Upper bound for the `np.percentile` threshold, used for rescaling
        the intensity. Defaults to 98.
        threshold (float, optional): The lower bound for pixel values used to create a binary mask.
        Defaults to 0.35.
        wavelet (str): The type of wavelet to perform and can be any of the options
        `pywt.wavelist` outputs. Defaults to "db2".
        mode (str): An optional argument to choose the type of denoising performed. Its noted that
        choosing soft thresholding given additive noise finds the best approximation of the
        original image. Defaults to "soft".
        rescale_sigma (bool): If False, no rescaling of the user-provided `sigma` will be
        performed. The default of `True` rescales `sigma` appropriately if the image is rescaled
        internally. Defaults to "True".


    Returns:
        np.ndarray: The binary mask containing all of the candidate strokes.
    """
    input_image = restoration.denoise_wavelet(
        input_image, wavelet=wavelet, mode=mode, rescale_sigma=rescale_sigma
    )
    # Rescale the intensity using percentile ranges
    pmin_v, pmax_v = np.percentile(input_image, (pmin, pmax))
    input_image = exposure.rescale_intensity(input_image, in_range=(pmin_v, pmax_v))

    # Laplace filter to get the streaks
    input_image = filters.laplace(input_image, ksize=3)
    input_image = exposure.rescale_intensity(input_image, out_range=(0, 1))

    # Smoothing
    input_image = filters.gaussian(input_image, sigma=(0, gaussian_sigma))  # (y, x)

    # Exposure Adjustments
    input_image = exposure.adjust_gamma(input_image, gamma=gamma, gain=gamma_gain)
    input_image = exposure.adjust_log(input_image, gain=log_gain, inv=True)
    input_image = exposure.rescale_intensity(input_image, out_range=(0, 1))

    # apply threshold
    binary_mask = input_image > threshold

    return binary_mask


def _make_mask_dataframe(streak_data: StreakData, min_length: int = 70) -> None:
    """Converts the binary mask created by `_make_binary_mask` into a dataframe for
    processing. The streaks are labeled, pixel information (min_row, min_col, max_row, max_col)
    is evaluated and streak lengths / areas are calculated. In addition the `min_length` argument
    allows the user to filter out streaks shorter than it.

    Args:
        streak_data (StreakData): An instance of the StreakData Dataclass, holds all necessary
        data for streak correction.
        min_length (int): The lower threshold for filtering streaks in pixels. Defaults to 70.
        eccentricity_value (float): The threshold to filter out eccentricities lower than this
    """
    # Label all the candidate streaks
    labeled_streaks = measure.label(streak_data.streak_mask, connectivity=2, return_num=False)

    # if streaks detected, filter dataframe
    if len(np.unique(labeled_streaks)) > 1:
        # Gather properties of all the candidate streaks using regionprops.
        region_properties = measure.regionprops_table(
            label_image=labeled_streaks,
            cache=True,
            properties=[
                "label",
                "bbox",
                "eccentricity",
                "area",
            ],
        )

        # Convert dictionary of region properties to DataFrame.
        streak_data.streak_df = pd.DataFrame(region_properties)

        # Rename the bounding box columns.
        streak_data.streak_df.rename(
            {
                "bbox-0": "min_row",
                "bbox-1": "min_col",
                "bbox-2": "max_row",
                "bbox-3": "max_col",
                "area": "length",
            },
            axis="columns",
            inplace=True,
        )
        # Give the index column a name.
        streak_data.streak_df.index.names = ["index"]

        # Filter out eccentricities that are less than 0.99999 (only keep straight lines)
        # Filter out small areas (small lines)
        eccentricity_value = 0.9999999
        streak_data.filtered_streak_df = streak_data.streak_df.query(
            f"eccentricity > {eccentricity_value} and length > {min_length}"
        )
    else:
        # otherwise, make a blank df
        blank_df = pd.DataFrame({"min_row": [], "min_col": [], "max_row": [], "max_col": []})
        streak_data.filtered_streak_df = blank_df


def _make_filtered_mask(streak_data: StreakData) -> None:
    """Visualization Utility. Uses the filtered streak dataframe to create a binary mask, where
    1 indicates the pixels that will get corrected. This mask can be later saved and used for
    visualization purposes.

    Args:
        streak_data (StreakData): An instance of the StreakData Dataclass, holds all necessary
        data for streak correction.
    """
    streak_data.filtered_streak_mask = np.zeros(shape=streak_data.shape, dtype=np.uint8)
    for region in streak_data.filtered_streak_df.itertuples():
        streak_data.filtered_streak_mask[
            region.min_row : region.max_row, region.min_col : region.max_col
        ] = 1


def _make_box_outline(streak_data: StreakData) -> None:
    """Visualization Utility. Creates a box outline for each binary streak using the filtered
    streak dataframe. Outlines the streaks that will get corrected. This mask can be later saved
    and used for visualization purposes.

    Args:
        streak_data (StreakData): An instance of the StreakData Dataclass, holds all necessary
        data for streak correction.
    """
    padded_image = np.pad(
        np.zeros(shape=streak_data.shape, dtype=np.uint8), pad_width=(1, 1), mode="edge"
    )
    for region in streak_data.filtered_streak_df.itertuples():
        y, x = draw.rectangle_perimeter(
            start=(region.min_row + 1, region.min_col + 1),
            end=(region.max_row, region.max_col),
            clip=True,
            shape=streak_data.shape,
        )
        padded_image[y, x] = 1
    streak_data.boxed_streaks = util.crop(padded_image, crop_width=(1, 1))


def _make_correction_mask(streak_data: StreakData) -> None:
    """Visualization Utility. Creates the correction mask for each binary streak using the
    filtered streak DataFrame. Marks pixels which will be used for the correction method.
    This mask can be later saved and used for visualization purposes.

    Args:
        streak_data (StreakData): An instance of the StreakData Dataclass, holds all necessary
        data for streak correction.
    """
    padded_image = np.pad(
        np.zeros(shape=streak_data.shape, dtype=np.uint8), pad_width=(1, 1), mode="edge"
    )

    for region in streak_data.filtered_streak_df.itertuples():
        padded_image[region.min_row, region.min_col + 1 : region.max_col + 1] = np.ones(
            shape=(region.max_col - region.min_col)
        )
        padded_image[region.max_row + 1, region.min_col + 1 : region.max_col + 1] = np.ones(
            shape=(region.max_col - region.min_col)
        )

    streak_data.corrected_streak_mask = util.crop(padded_image, crop_width=(1, 1))


def _correct_streaks(streak_data: StreakData, input_image: np.ndarray) -> np.ndarray:
    """Corrects the streaks for the input image. Uses masks in the streak_data Dataclass.
    Performs the correction by averaging the pixels above and below the streak.

    Args:
        streak_data (StreakData): An instance of the StreakData Dataclass, holds all necessary
        data for streak correction.
        input_image (np.ndarray): The channel which is being corrected.

    Returns:
        np.ndarray: The corrected image.
    """
    # Pad the image for edge cases.
    padded_image = np.pad(input_image.copy(), pad_width=(1, 1), mode="edge")
    corrected_image = padded_image.copy()
    # Correct each streak
    for region in streak_data.filtered_streak_df.itertuples():
        corrected_image[region.max_row, region.min_col : region.max_col] = _correct_mean_alg(
            padded_image,
            region.min_row,
            region.max_row,
            region.min_col,
            region.max_col,
        )
    # Crop and return the 'unpadded' image.
    return util.crop(corrected_image, crop_width=(1, 1), copy=True)


def _correct_mean_alg(
    input_image: np.ndarray, min_row: int, max_row: int, min_col: int, max_col: int
) -> np.ndarray:
    """Performs streak-wise correction by: setting the value of each pixel in the streak to the
    mean of pixel above and below it.

    Args:
        input_image (np.ndarray): The channel to be corrected.
        min_row (int): The minimum row index of the streak. The y location where the streak
        starts.
        max_row (int): The maximum row index of the streak. The y location where the streak ends.
        min_col (int): The minimum column index of the streak. The x location where the streak
        starts.
        max_col (int): The maximum column index of the streak. The x location where the streak
        ends.

    Returns:
        np.ndarray: Returns the corrected streak.
    """
    streak_corrected: np.ndarray = np.mean(
        [
            # Row above
            input_image[min_row, min_col + 1 : max_col + 1],
            # Row below
            input_image[max_row + 1, min_col + 1 : max_col + 1],
        ],
        axis=0,
        dtype=input_image.dtype,
    )

    return streak_corrected


def save_corrected_channels(
    streak_data: StreakData,
    corrected_channels: xr.DataArray,
    data_dir: Path,
    save_streak_data=False,
) -> None:
    """Saves the corrected images in a subdirectory of `fov_dir`.

    Args:
        streak_data (StreakData): An instance of the StreakData Dataclass, holds all necessary
        data for streak correction.
        corrected_channels (xr.DataArray): The DataArray continaing the set of corrected channels.
        data_dir (Path): A directory containing the fov and all it's channels for correction.
        save_streak_data (bool): Saves the binary masks and dataframes contained in StreakData.

    """
    # Create the directory to store the corrected tiffs
    streak_data.corrected_dir = Path(data_dir, streak_data.fov + "-corrected")
    if not os.path.exists(streak_data.corrected_dir):
        os.makedirs(streak_data.corrected_dir)

    # Save the corrected tiffs
    for channel in corrected_channels.channels.values:
        img: np.ndarray = corrected_channels.loc[:, :, channel].values
        fp = Path(streak_data.corrected_dir, channel + ".tiff")
        save_image(fp, img)

    # Save streak masks
    if save_streak_data:
        _save_streak_masks(streak_data=streak_data)


def streak_correction(
    fov_data: xr.DataArray,
    streak_channel: str = "Noodle",
    visualization_masks: bool = False,
) -> Tuple[xr.DataArray, StreakData]:
    """Takes an DataArray representation of a fov and a user specified image for streak detection.
    Once all the streaks have been detected on that image, they are corrected via an averaging
    method. The function can also returns a DataClass containing various binary masks and
    dataframes which were used for filtering and correction when `visualization_masks` is True.

    Args:
        fov_data (xr.DataArray): The data structure containing all of the channels to be processed
        and corrected.
        streak_channel (str, optional): The name of the channel used (without the file extension)
        for identifying the streaks. Defaults to "Noodle".
        visualization_masks (bool, optional): If `True`, adds binary masks for visualization to
        the StreakData Dataclass which gets returned. Defaults to "False".

    Returns:
        Tuple[xr.DataArray, StreakData]: A tuple of the DataArray housing the corrected images,
        and the streak data containing masks and dataframes for analysis and visualization.
    """
    # Initialize the streak DataClass
    streak_data = StreakData()
    streak_data.streak_channel = streak_channel
    streak_data.fov = fov_data.fovs.values[0]

    fov_data = fov_data[0, ...]

    #  Get the correct channel for mask generation.
    with fov_data.loc[:, :, streak_channel] as channel_image:
        streak_data.shape = channel_image.shape
        # Create and filter the binary masks
        streak_data.streak_mask = _make_binary_mask(input_image=channel_image)
        _make_mask_dataframe(streak_data=streak_data)

    # Get the file names.
    channel_fn = fov_data.channels.values.tolist()

    # Initialize the corrected image fov dimensions.
    fov_dim_size: int = len(channel_fn)
    row_size, col_size = streak_data.shape
    cor_img_data = np.zeros(shape=(row_size, col_size, fov_dim_size), dtype=fov_data.dtype)

    # Correct streaks and add them to the np.array
    for idx, channel in enumerate(fov_data.channels.values):
        input_channel = fov_data.loc[:, :, channel]
        cor_img_data[:, :, idx] = _correct_streaks(
            streak_data=streak_data, input_image=input_channel
        )

    # Create xarray from np.array
    corrected_channels = xr.DataArray(
        data=cor_img_data,
        coords=[range(row_size), range(col_size), fov_data.channels.values],
        dims=["rows", "cols", "channels"],
    )

    # Add mask information / visualization masks
    if visualization_masks:
        _make_box_outline(streak_data=streak_data)
        _make_correction_mask(streak_data=streak_data)
        _make_filtered_mask(streak_data=streak_data)

    return (corrected_channels, streak_data)



import copy
import json
import os
import warnings
from collections import Counter

import pandas as pd
from alpineer import io_utils


def rename_missing_fovs(fov_data):
    """Identify FOVs that are missing the 'name' key and create one with value placeholder_{n}.

    Args:
        fov_data (dict): the FOV run JSON data.

    Returns:
        dict: a copy of the run JSON data with placeholder names for FOVs that lack one
    """
    copy_fov_data = copy.deepcopy(fov_data)

    # count of FOVs that are missing the 'name' key
    missing_count = 0

    # iterate over each FOV and add a placeholder name if necessary
    for fov in copy_fov_data["fovs"]:
        if "name" not in fov.keys():
            missing_count += 1
            fov["name"] = f"placeholder_{missing_count}"

    return copy_fov_data


def rename_duplicate_fovs(tma_fovs):
    """Identify and rename duplicate FOV names in `fov_list` For a given FOV name, the subsequent
       duplicates get renamed `{FOV}_duplicate{n}`.

    Args:
        tma_fovs (dict):
            The TMA run JSON, should contain a `'fovs'` key defining the list of FOVs

    Returns:
        dict:
            The same run JSON with the FOVs renamed to account for duplicates
    """
    # used for identifying the number of times each FOV was found
    fov_count = {}

    # iterate over each FOV
    for fov in tma_fovs["fovs"]:
        if fov["name"] not in fov_count:
            fov_count[fov["name"]] = 0

        fov_count[fov["name"]] += 1

        if fov_count[fov["name"]] > 1:
            fov["name"] = "%s_duplicate%d" % (fov["name"], fov_count[fov["name"]] - 1)

    return tma_fovs


def list_moly_fovs(bin_file_dir, fov_list=None):
    """Lists all of the FOVs in a directory or provided list which are moly FOVs.

    Args:
        bin_file_dir (str): path to bin files
        fov_list (list): list of fov names to check, default None will check all fovs in dir

    Returns:
        list: list of FOVs which are moly FOVs
    """
    run_name = os.path.basename(bin_file_dir)

    # check provided fovs
    if fov_list:
        json_files = [fov + ".json" for fov in fov_list]
    # check all fovs in bin_file_dir
    else:
        json_files = io_utils.list_files(bin_file_dir, ".json")
        # ignore run file
        if run_name + ".json" in json_files:
            json_files.remove(run_name + ".json")
    moly_fovs = []

    for file in json_files:
        json_path = os.path.join(bin_file_dir, file)
        json_file = read_json_file(json_path, encoding="utf-8")

        if json_file.get("standardTarget", "") == "Molybdenum Foil":
            moly_name = file.split(".json")[0]
            moly_fovs.append(moly_name)

    return moly_fovs


def read_json_file(json_path, encoding=None):
    """Reads json file and returns json file object while verifying dirs exist.

    Args:
        json_path (str): path to json file
        encoding (str): type of file encoding
    Returns:
        dict: the JSON object loaded from `json_path`
    """
    # call to validate paths will raise errors if anything wrong, and do nothing if
    # file path valid
    io_utils.validate_paths(json_path)

    with open(json_path, mode="r", encoding=encoding) as jp:
        json_file = json.load(jp)

    return json_file


def write_json_file(json_path, json_object, encoding=None):
    """Writes json file object to json file. Raises error if directory doesn't exist.

    Args:
        json_path (str): full path to write json file
        json_object (dict): data to save to the file
        encoding (str): type of file encoding
    """
    # get the path minus the proposed file name.
    dir_path = os.path.dirname(os.path.abspath(json_path))

    # Raises error if path doesnt exist
    io_utils.validate_paths(dir_path)

    with open(json_path, mode="w", encoding=encoding) as jp:
        json.dump(json_object, jp)


def split_run_file(run_dir, run_file_name, file_split: list):
    """Splits a run json file into smaller fov amount files as defined by the user.

    Args:
        run_dir (str): path to directory containing the run file
        run_file_name (str): name of the run file to split
        file_split (list): list of ints defining how to break up the fovs into new jsons
    """
    json_path = os.path.join(run_dir, run_file_name)
    full_json = read_json_file(json_path, encoding="utf-8")

    # check list is valid FOV split
    if not sum(file_split) == len(full_json["fovs"]):
        raise ValueError(
            "Sum of the provided list does not match the number of FOVs in the run file.\n"
            f"list sum: {sum(file_split)}, total FOVs in the JSON: {len(full_json['fovs'])}"
        )

    # split the run json into smaller files and save to run_dir
    start = 0
    for i in range(0, len(file_split)):
        json_i = copy.deepcopy(full_json)
        stop = start + file_split[i]
        json_i["fovs"] = json_i["fovs"][start:stop]
        start = start + file_split[i]

        save_path = os.path.join(
            run_dir, run_file_name.split(".json")[0] + "_part" + str(i + 1) + ".json"
        )
        write_json_file(save_path, json_i, encoding="utf-8")


def check_for_empty_files(bin_file_dir):
    """Check for any empty json files and warn the user.

    Args:
        bin_file_dir (str): directory containing the bin and json files

    Return:
        list: fov files with empty json, if none returns empty list
        raises a warning
    """
    # retrieve all fovs in bin_file_dir
    fov_names = io_utils.remove_file_extensions(io_utils.list_files(bin_file_dir, substrs=".bin"))
    empty_json_files = []

    # check each json file for size 0
    for fov in fov_names:
        fov_path = os.path.join(bin_file_dir, fov + ".json")
        if os.path.getsize(fov_path) == 0:
            empty_json_files.append(fov)

    # print a warning to the user when there are empty files
    if empty_json_files:
        warnings.warn(
            (
                "The following FOVs have empty json files and will not be processed:"
                f"\n {empty_json_files}"
            ),
            UserWarning,
        )

    # return the list of fov names
    return empty_json_files


def check_fov_resolutions(bin_file_dir, run_name, save_path=None):
    """Use the run metadata to calculate the resolution of each fov.

    Args:
        bin_file_dir (str): directory containing the run json file
        run_name (str): name of the run and corresponding run file
        save_path (str): path to save data to, default None doesn't write data out

    Returns:
        pd.DataFrame: details fov names and corresponding resolution value
    """
    # read in run metadata
    run_file_path = os.path.join(bin_file_dir, run_name + ".json")
    io_utils.validate_paths([run_file_path])
    run_metadata = read_json_file(run_file_path, encoding="utf-8")

    fov_names, custom_names, resolutions = [], [], []
    for fov in run_metadata.get("fovs", ()):
        # ignore moly resolutions
        if fov.get("standardTarget") == "Molybdenum Foil":
            continue

        # get fov names
        fov_number = fov.get("runOrder")
        default_name = f"fov-{fov_number}-scan-1"
        custom_name = fov.get("name")
        fov_names.append(default_name)
        custom_names.append(custom_name)

        # retrieve pixel and micron specs
        fov_pixel_length = fov.get("frameSizePixels")["width"]
        fov_micron_length = fov.get("fovSizeMicrons")

        # calculate and save fov resolution per 400 microns
        mult = 400 / fov_micron_length
        pixels_adj = int(mult * fov_pixel_length)

        # save values
        resolutions.append(pixels_adj)

    resolution_data = pd.DataFrame(
        {"fov": fov_names, "name": custom_names, "pixels / 400 microns": resolutions}
    )

    if len(set(resolution_data["pixels / 400 microns"])) > 1:
        val = Counter(resolutions).most_common(1)[0][0]
        print(
            "Inconsistent resolutions.\n"
            f"All FOVs are {val} pixels per 400 microns except the following:"
        )

        # print resolutions that are in the minority
        subset = resolution_data[resolution_data["pixels / 400 microns"] != val]
        print(subset.to_string(index=False))
    else:
        print(f"FOV resolutions are all {pixels_adj} pixels / 400 microns")

    if save_path:
        resolution_data.to_csv(save_path, index=False)

    return resolution_data


def missing_fov_check(bin_file_dir, run_name):
    """Use the run metadata to check if all FOV data was generated.

    Args:
        bin_file_dir (str): directory containing the run json file
        run_name (str): name of the run and corresponding run file

    Raises:
        Warning if any FOVs specified in the run file doesn't have associated bin/json files
    """
    # read in run file
    run_file_path = os.path.join(bin_file_dir, run_name + ".json")
    io_utils.validate_paths([run_file_path])
    run_metadata = read_json_file(run_file_path, encoding="utf-8")

    missing_fovs = {}
    for fov in run_metadata.get("fovs", ()):
        if fov.get("standardTarget") == "Molybdenum Foil":
            continue

        # get fov names
        fov_number = fov.get("runOrder")
        default_name = f"fov-{fov_number}-scan-1"
        custom_name = fov.get("name")

        # add default name and custom name for missing data
        if not os.path.exists(
            os.path.join(bin_file_dir, default_name + ".bin")
        ) or not os.path.exists(os.path.join(bin_file_dir, default_name + ".json")):
            missing_fovs[default_name] = [custom_name]

        elif os.path.getsize(os.path.join(bin_file_dir, default_name + ".json")) == 0:
            missing_fovs[default_name] = [custom_name]

    if missing_fovs:
        missing_fovs = pd.DataFrame(missing_fovs, index=[0]).T
        missing_fovs.columns = ["fov_name"]
        warnings.warn(
            "The following FOVs were not processed due to missing/empty/late files: \n"
            f"{missing_fovs}"
        )



# adapted from https://machinelearningmastery.com/curve-fitting-with-python/
import copy
import glob
import json
import os
import pathlib
import warnings
from typing import Literal, Union

import matplotlib.pyplot as plt
import natsort as ns
import numpy as np
import pandas as pd
from alpineer import image_utils, io_utils, load_utils, misc_utils
from mibi_bin_tools.bin_files import extract_bin_files, get_median_pulse_height
from mibi_bin_tools.panel_utils import make_panel
from scipy.optimize import curve_fit

from toffy.json_utils import check_for_empty_files, read_json_file, write_json_file


def write_counts_per_mass(base_dir, output_dir, fov, masses, start_offset=0.5, stop_offset=0.5):
    """Records the total counts per mass for the specified FOV.

    Args:
        base_dir (str): the directory containing the FOV
        output_dir (str): the directory where the csv file will be saved
        fov (str): the name of the fov to extract
        masses (list): the list of masses to extract counts from
        start_offset (float): beginning value for integrating mass peak
        stop_offset (float): ending value for integrating mass peak
    """
    # create panel with extraction criteria
    panel = make_panel(mass=masses, low_range=start_offset, high_range=stop_offset)

    array = extract_bin_files(
        data_dir=base_dir,
        out_dir=None,
        include_fovs=[fov],
        panel=panel,
        intensities=False,
    )
    # we only care about pulse counts, not intensities
    array = array.loc[fov, "pulse", :, :, :]
    channel_count = np.sum(array, axis=(0, 1))

    # create df to hold output
    fovs = np.repeat(fov, len(masses))
    out_df = pd.DataFrame({"mass": masses, "fov": fovs, "channel_count": channel_count})
    out_df.to_csv(os.path.join(output_dir, fov + "_channel_counts.csv"), index=False)


def write_mph_per_mass(
    base_dir, output_dir, fov, masses, start_offset=0.5, stop_offset=0.5, proficient=False
):
    """Records the median pulse height (MPH) per mass for the specified FOV.

    Args:
        base_dir (str): the directory containing the FOV
        output_dir (str): the directory where the csv file will be saved
        fov (str): the name of the fov to extract
        masses (list): the list of masses to extract MPH from
        start_offset (float): beginning value for calculating mph values
        stop_offset (float): ending value for calculating mph values
        proficient (bool): whether proficient MPH data is written or not
    """
    # compute pulse heights
    panel = make_panel(
        mass=masses, target_name=masses, low_range=start_offset, high_range=stop_offset
    )

    # generate the MPH values for each mass
    mph_vals = list(
        get_median_pulse_height(data_dir=base_dir, fov=fov, channels=masses, panel=panel).values()
    )

    fovs = np.repeat(fov, len(masses))
    out_df = pd.DataFrame({"mass": masses, "fov": fovs, "pulse_height": mph_vals})
    pulse_heights_file = (
        fov + "_pulse_heights_proficient.csv" if proficient else fov + "_pulse_heights.csv"
    )
    out_df.to_csv(os.path.join(output_dir, pulse_heights_file), index=False)


def create_objective_function(obj_func):
    """Creates a function of specified type to be used for fitting a curve.

    Args:
        obj_func (str): the desired objective function. Must be either poly_2, ..., poly_5, or log

    Returns:
        Callable: the function which will be optimized
    """
    # input validation
    valid_funcs = ["poly_2", "poly_3", "poly_4", "poly_5", "log", "exp"]
    if obj_func not in valid_funcs:
        raise ValueError("Invalid function, must be one of {}".format(valid_funcs))

    # define objective functions
    def poly_2(x, a, b, c):
        return a * x + b * x**2 + c

    def poly_3(x, a, b, c, d):
        return a * x + b * x**2 + c * x**3 + d

    def poly_4(x, a, b, c, d, e):
        return a * x + b * x**2 + c * x**3 + d * x**4 + e

    def poly_5(x, a, b, c, d, e, f):
        return a * x + b * x**2 + c * x**3 + d * x**4 + e * x**5 + f

    def log(x, a, b):
        # edge case appears when x is a single int/float that returns non NaN/inf for np.log
        # in this case, return the base np.log calculation
        try:
            return (a * np.ma.log(x) + b).filled(0)
        except AttributeError:
            return a * np.log(x) + b

    def exp(x, a, b, c, d):
        try:
            x_log = (np.ma.log(x)).filled(0)
        except AttributeError:
            x_log = a * np.log(x) + b

        return a * x_log + b * x_log**2 + c * x_log**3 + d

    objectives = {
        "poly_2": poly_2,
        "poly_3": poly_3,
        "poly_4": poly_4,
        "poly_5": poly_5,
        "log": log,
        "exp": exp,
    }

    return objectives[obj_func]


def fit_calibration_curve(
    x_vals,
    y_vals,
    obj_func,
    outliers=None,
    plot_fit=False,
    save_path=None,
    x_label=None,
    y_label=None,
    title=None,
    show_plot=False,
):
    """Finds the optimal weights to fit the supplied values for the specified function.

    Args:
        x_vals (list): the x values to be fit
        y_vals (list): the y value to be fit
        obj_func (str): the name of the function that will be fit to the data
        outliers (tuple or None): optional tuple of ([x_coords], [y_coords]) to plot
        plot_fit (bool): whether or not to plot the fit of the function vs the values
        save_path (str or None): location to save the plot of the fitted values
        x_label (str or None): label for the x-axis
        y_label (str or None):label for the y-axis
        title (str or None): label for the plot title
        show_plot (bool): whether to show plot, default False

    Returns:
        list: the weights of the fitted function
    """
    # get objective function
    objective = create_objective_function(obj_func)

    # get fitted values
    popt, _ = curve_fit(objective, x_vals, y_vals)

    # plot overlay of predicted fit and real values
    if plot_fit:
        plt.scatter(x_vals, y_vals)
        x_line = np.arange(min(x_vals), max(x_vals), 1)
        y_line = objective(x_line, *popt)
        plt.plot(x_line, y_line, "--", color="red")

        # add labels
        if x_label:
            plt.xlabel(x_label)
        if y_label:
            plt.ylabel(y_label)
        if title:
            plt.title(title)

        if outliers is not None:
            plt.scatter(outliers[0], outliers[1])

        if save_path is not None:
            plt.savefig(save_path)

        if show_plot:
            plt.show()
        plt.close()

    return popt


def create_prediction_function(name, weights):
    """Creates a prediction function given a specified function type and fitted weights.

    Args:
        name (str): name of the function to use
        weights (list): list of fitted weights

    Returns:
        func: prediction function
    """
    # get function based on specified pred type
    obj_func = create_objective_function(name)

    # define function which takes only x as input, uses weights for remaining parameters
    def pred_func(x):
        output = obj_func(x, *weights)
        return output

    return pred_func


def combine_run_metrics(run_dir, substring, warn_overwrite=True):
    """Combines the specified metrics from different FOVs into a single file.

    Args:
        run_dir (str): the directory containing the files
        substring(str): the substring contained within the files to be combined
        warn_overwrite (bool): whether to warn if existing `_combined.csv` file found
    """
    # retrieve all pulse height files, but ignore anything mass proficient
    files = io_utils.list_files(run_dir, substring)
    files = [file for file in files if "_proficient" not in file]

    # validate inputs
    if len(files) == 0:
        raise ValueError("No files found in {}".format(run_dir))

    if substring + "_combined.csv" in files:
        if warn_overwrite:
            warnings.warn(
                "Removing previously generated combined {} file in {}".format(substring, run_dir)
            )
        os.remove(os.path.join(run_dir, substring + "_combined.csv"))
        files = [file for file in files if "combined" not in file]

    # collect all metrics files
    metrics = []
    for file in files:
        metrics.append(pd.read_csv(os.path.join(run_dir, file)))

    # check that all files are the same length
    if len(metrics) > 1:
        base_len = len(metrics[0])
        for i in range(1, len(metrics)):
            if len(metrics[i]) != base_len:
                raise ValueError(
                    "Not all {} files are the same length: file {} does not matchfile {}".format(
                        substring, files[0], files[i]
                    )
                )

    metrics = pd.concat(metrics)

    metrics.to_csv(os.path.join(run_dir, substring + "_combined.csv"), index=False)


def combine_tuning_curve_metrics(dir_list, count_range):
    """Combines metrics together into a single dataframe for fitting a turning curve.

    Args:
        dir_list (list): list of directories to pull metrics from
        count_range (tuple): range of appropriate mass count values for the FOVs

    Returns:
        pd.DataFrame: dataframe containing aggregates metrics
    """
    # create list to hold all extracted data
    all_dirs, excluded_fovs = [], []
    # loop through each run folder
    for dir in dir_list:
        pulse_heights = pd.read_csv(os.path.join(dir, "fov-1-scan-1_pulse_heights.csv"))
        channel_counts = pd.read_csv(os.path.join(dir, "fov-1-scan-1_channel_counts.csv"))

        # check for extreme count values
        extreme_val = False
        if count_range:
            counts = channel_counts.channel_count
            if any(counts <= count_range[0]) or any(counts >= count_range[1]):
                extreme_val = True
                excluded_fovs.append(os.path.basename(dir))
        # do not add extreme value fovs to the combined data frame
        if extreme_val:
            continue

        # combine tables together
        combined = pulse_heights.merge(channel_counts, "outer", on=["fov", "mass"])
        if len(combined) != len(pulse_heights):
            raise ValueError(
                "Pulse heights and channel counts must be generated for the same "
                "mass ranges and fovs. However, the data the following does do not "
                "exactly match: {}".format(dir)
            )

        # add directory label and add to list
        combined["directory"] = dir
        all_dirs.append(combined)

    if len(excluded_fovs) > 0:
        excluded_fovs = ns.natsorted(excluded_fovs)
        print(
            (
                "The counts for the FOV contained in the following folders are outside of "
                "the expected range and will be excluded: "
            ),
            *excluded_fovs,
            sep="\n- ",
        )
    elif count_range:
        print("No extreme values detected.")

    # combine data from each dir together
    all_data = pd.concat(all_dirs)
    all_data.reset_index()

    # check for sufficient data
    if len(set(all_data.directory)) < 4:
        raise ValueError("Invalid amount of FOV data. Please choose another sweep.")

    # create normalized counts column
    subset = all_data[["channel_count", "mass"]]
    all_data["norm_channel_count"] = subset.groupby("mass").transform(lambda x: (x / x.max()))

    return all_data


def plot_voltage_vs_counts(sweep_fov_paths, combined_data, save_path):
    """Creates a barplot of voltage and maximum channel counts and saves the image.

    Args:
        sweep_fov_paths (list): paths to the sweep folders
        combined_data (pd.DataFrame): combined data frame of pulse height and channel info
        save_path (str): where to save plot to
    """
    voltages, max_channel_counts = [], []

    # get voltage and max channel count for each fov
    for fov_path in sweep_fov_paths:
        fov_data = read_json_file(os.path.join(fov_path, "fov-1-scan-1.json"))

        # locate index storing the detector voltage
        for j in range(0, len(fov_data["hvDac"])):
            if fov_data["hvDac"][j]["name"] == "Detector":
                index = j
                break
        fov_voltage = fov_data["hvDac"][index]["currentSetPoint"]
        fov_counts = combined_data.loc[combined_data["directory"] == fov_path, "channel_count"]

        voltages.append(fov_voltage)
        max_channel_counts.append(fov_counts.max() / 1000000)

    plt.bar(voltages, max_channel_counts)
    plt.xlabel("Voltage")
    plt.ylabel("Maximum Channel Counts (in millions)")
    plt.savefig(save_path)
    plt.close()


def show_multiple_plots(rows, cols, image_paths, image_size=(17, 12)):
    """Displays a single image with multiple plots.

    Args:
        rows (int): number of rows of plot grid
        cols (int): number of columns of plot grid
        image_paths (list): list of paths to the previously saved plot images
        image_size (tuple): length and width of the new image produced
    """
    fig = plt.figure(figsize=image_size)

    # add each plot to the figure
    for img, path in enumerate(image_paths):
        plot = plt.imread(path)
        fig.add_subplot(rows, cols, img + 1)
        plt.imshow(plot)
        plt.axis("off")

    plt.show()


def create_tuning_function(
    sweep_path,
    moly_masses=[92, 94, 95, 96, 97, 98, 100],
    save_path=os.path.join("..", "tuning_curves", "sample_tuning_curve.json"),
    count_range=(0, 3000000),
):
    """Creates a tuning curve for an instrument based on the provided moly sweep.

    Args:
        sweep_path (str): path to folder containing a detector sweep
        moly_masses (list): list of masses to use for fitting the curve
        save_path (str): path to save the weights of the tuning curve
        count_range (tuple): range of appropriate mass count values for the FOVs
    """
    # get all folders from the sweep
    sweep_fovs = io_utils.list_folders(sweep_path)
    sweep_fov_paths = [os.path.join(sweep_path, fov) for fov in sweep_fovs]
    # check for sufficient directory structure
    if len(sweep_fovs) < 4:
        raise ValueError("Invalid amount of FOV folders. Please use at least 4 voltages.")

    # compute pulse heights and channel counts for each FOV if files don't already exist
    for fov_path in sweep_fov_paths:
        if not os.path.exists(
            os.path.join(fov_path, "fov-1-scan-1_pulse_heights.csv")
        ) or not os.path.exists(os.path.join(fov_path, "fov-1-scan-1_channel_counts.csv")):
            # check for bin file in each folder
            bin_file = io_utils.list_files(fov_path, substrs="bin")
            if len(bin_file) == 0:
                raise ValueError(f"No bin file detected in {fov_path}")

            write_mph_per_mass(
                base_dir=fov_path,
                output_dir=fov_path,
                fov="fov-1-scan-1",
                masses=moly_masses,
            )
            write_counts_per_mass(
                base_dir=fov_path,
                output_dir=fov_path,
                fov="fov-1-scan-1",
                masses=moly_masses,
            )

    if count_range:
        # combine all data together into single df for comparison
        all_data = combine_tuning_curve_metrics(sweep_fov_paths, count_range=None)

        # generate curve with extreme values included
        fit_calibration_curve(
            all_data["pulse_height"].values,
            all_data["norm_channel_count"].values,
            "exp",
            plot_fit=True,
            save_path=os.path.join(sweep_path, "function_fit_all_data.jpg"),
            x_label="Median Pulse Height",
            y_label="Normalized Channel Counts",
            title="Tuning Curve (all data)",
            show_plot=False,
        )

        # plot voltage against maximum channel count
        plot_voltage_vs_counts(
            sweep_fov_paths,
            all_data,
            save_path=os.path.join(sweep_path, "voltage_vs_counts.jpg"),
        )

        show_multiple_plots(
            1,
            2,
            [
                os.path.join(sweep_path, "function_fit_all_data.jpg"),
                os.path.join(sweep_path, "voltage_vs_counts.jpg"),
            ],
        )

    # combine tuning date into single df, if count_range given then extreme values excluded
    tuning_data = combine_tuning_curve_metrics(sweep_fov_paths, count_range=count_range)

    # generate fitted curve
    tuning_coeffs = fit_calibration_curve(
        tuning_data["pulse_height"].values,
        tuning_data["norm_channel_count"].values,
        "exp",
        plot_fit=True,
        save_path=os.path.join(sweep_path, "function_fit.jpg"),
        x_label="Median Pulse Height",
        y_label="Normalized Channel Counts",
        title="Tuning Curve",
        show_plot=True,
    )

    # save the fitted curve
    norm_json = {"name": "exp", "weights": tuning_coeffs.tolist()}
    write_json_file(json_path=save_path, json_object=norm_json)


def identify_outliers(x_vals, y_vals, obj_func, outlier_fraction=0.1):
    """Finds the indices of outliers in the provided data to prune for subsequent curve fitting.

    Args:
        x_vals (np.array): the x values of the data being analyzed
        y_vals (np.array): the y values of the data being analyzed
        obj_func (str): the objective function to use for curve fitting to determine outliers
        outlier_fraction (float): the fractional deviation from predicted value required in
            order to classify a data point as an outlier

    Returns:
        np.array: the indices of the identified outliers
    """
    # get objective function
    objective = create_objective_function(obj_func)

    # get fitted values
    popt, _ = curve_fit(objective, x_vals, y_vals)

    # create generate function
    func = create_prediction_function(name=obj_func, weights=popt)

    # generate predictions
    preds = func(x_vals)

    # specify outlier bounds based on multiple of predicted value
    upper_bound = preds * (1 + outlier_fraction)
    lower_bound = preds * (1 - outlier_fraction)

    # identify outliers
    outlier_mask = np.logical_or(y_vals > upper_bound, y_vals < lower_bound)
    outlier_idx = np.where(outlier_mask)[0]

    return outlier_idx


def smooth_outliers(vals, outlier_idx, smooth_range=2):
    """Performs local smoothing on the provided outliers.

    Args:
        vals (np.array): the complete list of values to be smoothed
        outlier_idx (np.array): the indices of the outliers in *vals* argument
        smooth_range (int): the number of adjacent values in each direction to use for smoothing

    Returns:
        np.array: the smoothed version of the provided vals
    """
    smoothed_vals = copy.deepcopy(vals)
    vals = np.array(vals)

    for outlier in outlier_idx:
        previous_vals = smoothed_vals[(outlier - smooth_range) : outlier]

        if outlier == len(vals):
            # last value in list, can't average using subsequent values
            subsequent_vals = []
        else:
            # not the last value, we can use remaining values to get an estimate
            subsequent_indices = np.arange(outlier + 1, len(vals))
            valid_subs_indices = [idx for idx in subsequent_indices if idx not in outlier_idx]
            subsequent_indices = np.array(valid_subs_indices)[:smooth_range]

            # check to make sure there are valid subsequent indices
            if len(subsequent_indices) > 0:
                subsequent_vals = vals[subsequent_indices]
            else:
                subsequent_vals = np.array([])

        new_val = np.mean(np.concatenate([previous_vals, subsequent_vals]))
        smoothed_vals[outlier] = new_val

    return smoothed_vals


def fit_mass_mph_curve(mph_vals, mass, save_dir, obj_func, min_obs=10, outlier_fraction=0.1):
    """Fits a curve for the MPH over time for the specified mass.

    Args:
        mph_vals (list): mph for each FOV in the run
        mass (str or int): the mass being fit
        save_dir (str): the directory to save the fit parameters
        obj_func (str): the function to use for constructing the fit
        min_obs (int): the minimum number of observations to fit a curve, otherwise uses median
        outlier_fraction (float): the fractional deviation from predicted value required in
            order to classify a data point as an outlier
    """
    fov_order = np.linspace(0, len(mph_vals) - 1, len(mph_vals))
    save_path = os.path.join(save_dir, str(mass) + "_mph_fit.jpg")

    if len(mph_vals) > min_obs:
        # find outliers in the MPH vals
        outlier_idx = identify_outliers(
            x_vals=fov_order, y_vals=mph_vals, obj_func=obj_func, outlier_fraction=outlier_fraction
        )

        # replace with local smoothing around that point
        smoothed_vals = smooth_outliers(vals=mph_vals, outlier_idx=outlier_idx)

        # if outliers identified, generate tuple to pass to plotting function
        if len(outlier_idx) > 0:
            outlier_x = fov_order[outlier_idx]
            outlier_y = mph_vals[outlier_idx]
            outlier_tup = (outlier_x, outlier_y)
        else:
            outlier_tup = None

        # fit curve
        weights = fit_calibration_curve(
            x_vals=fov_order,
            y_vals=smoothed_vals,
            obj_func=obj_func,
            outliers=outlier_tup,
            plot_fit=True,
            save_path=save_path,
        )
    else:
        # default to using the median instead for short runs with small number of FOVs
        mph_median = np.median(mph_vals)
        if obj_func == "poly_2":
            weight_len = 3
        elif obj_func == "poly_3":
            weight_len = 4
        else:
            raise ValueError("Unsupported objective function provided: {}".format(obj_func))

        # plot median
        plt.axhline(y=mph_median, color="r", linestyle="-")
        plt.plot(fov_order, mph_vals, ".")
        plt.savefig(save_path)
        plt.close()

        # all coefficients except intercept are 0
        weights = np.zeros(weight_len)
        weights[-1] = mph_median

    mass_json = {"name": obj_func, "weights": weights.tolist()}
    mass_path = os.path.join(save_dir, str(mass) + "_norm_func.json")

    write_json_file(json_path=mass_path, json_object=mass_json)


def create_fitted_mass_mph_vals(pulse_height_df, obj_func_dir):
    """Uses the mph curves for each mass to generate a smoothed mph estimate.

    Args:
        pulse_height_df (pd.DataFrame): contains the MPH value per mass for all FOVs
        obj_func_dir (str): directory containing the curves generated for each mass.
            Set to None if autogain flag is set for calling functions.

    Returns:
        pd.DataFrame: updated dataframe with fitted version of each MPH value for each mass
    """
    # get all masses
    masses = np.unique(pulse_height_df["mass"].values)

    # create column to hold fitted values
    pulse_height_df["pulse_height_fit"] = 0.0

    # create x axis values
    num_fovs = len(np.unique(pulse_height_df["fov"]))
    fov_order = np.linspace(0, num_fovs - 1, num_fovs)

    for mass in masses:
        mass_idx = pulse_height_df["mass"] == mass

        # if channel-specific prediction function does not exist, set to 0
        mass_path = os.path.join(obj_func_dir, str(mass) + "_norm_func.json")

        if not os.path.exists(mass_path):
            pulse_height_df.loc[mass_idx, "pulse_height_fit"] = 0.0
            continue

        # load channel-specific prediction function
        mass_json = read_json_file(mass_path)

        # compute predicted MPH
        name, weights = mass_json["name"], mass_json["weights"]
        pred_func = create_prediction_function(name=name, weights=weights)
        pred_vals = pred_func(fov_order)

        # update df
        pulse_height_df.loc[mass_idx, "pulse_height_fit"] = pred_vals

    return pulse_height_df


def create_fitted_pulse_heights_file(
    pulse_height_dir, panel_info, norm_dir, mass_obj_func, autogain=False, **kwargs
):
    """Create a single file containing the pulse heights after fitting a curve per mass.

    Args:
        pulse_height_dir (str): path to directory containing pulse height csvs
        panel_info (pd.DataFrame): the panel for this dataset
        norm_dir (str): the directory where normalized images will be saved
        mass_obj_func (str): the objective function used to fit the MPH over time per mass
        autogain (bool): whether the run had autogain turned on or not
        **kwargs (dict): additional parameters to pass to `fit_mass_mph_curve` for outlier detection

    Returns:
        pd.DataFrame: the combined pulse heights file
    """
    # create variables for mass fitting
    masses = panel_info["Mass"].values
    fit_dir = os.path.join(norm_dir, "curve_fits")
    os.makedirs(fit_dir)

    # combine fov-level files together
    combine_run_metrics(run_dir=pulse_height_dir, substring="pulse_heights")
    pulse_height_df = pd.read_csv(os.path.join(pulse_height_dir, "pulse_heights_combined.csv"))

    # order by FOV
    ordering = ns.natsorted((pulse_height_df["fov"].unique()))
    pulse_height_df["fov"] = pd.Categorical(
        pulse_height_df["fov"], ordered=True, categories=ordering
    )
    pulse_height_df = pulse_height_df.sort_values("fov")

    # loop over each mass, and fit a curve for MPH over the course of the run
    for mass in masses:
        mph_vals = pulse_height_df.loc[pulse_height_df["mass"] == mass, "pulse_height"].values

        # only create a _norm_func file if the mph_vals are non-zero
        if np.all(mph_vals == 0):
            warnings.warn("Skipping normalization for mass %s with all zero pulse heights" % mass)
            continue

        # for non-autogain runs, use the default min_obs at 10 for curve fitting
        # for autogain runs, the median of all observed MPH values should always be used
        min_obs = 10**10 if autogain else kwargs.get("min_obs", 10)
        outlier_fraction = kwargs.get("outlier_fraction", 0.1)

        fit_mass_mph_curve(
            mph_vals=mph_vals,
            mass=mass,
            save_dir=fit_dir,
            obj_func=mass_obj_func,
            min_obs=min_obs,
            outlier_fraction=outlier_fraction,
        )

    # update pulse_height_df to include fitted mph values
    pulse_height_df = create_fitted_mass_mph_vals(
        pulse_height_df=pulse_height_df, obj_func_dir=fit_dir
    )

    return pulse_height_df


def normalize_fov(img_data, norm_vals, norm_dir, fov, channels, extreme_vals):
    """Normalize a single FOV with provided normalization constants for each channel."""
    # create directory to hold normalized images
    output_fov_dir = os.path.join(norm_dir, fov)
    if os.path.exists(output_fov_dir):
        print("output directory {} already exists, data will be overwritten".format(output_fov_dir))
    else:
        os.makedirs(output_fov_dir)

    # cap maximum normalization increase at 10X, ignore 0 (skipped) norm_vals
    abnormal_increase_mask = np.where(np.logical_and(norm_vals < 0.1, norm_vals != 0))[0]
    if len(abnormal_increase_mask) > 0:
        abnormal_increase_chans = np.array(channels)[abnormal_increase_mask]
        warnings.warn(
            "The following channel(s) are below 10% sensitivity "
            "for fov {}. Normalization capped at 10X for: {}. "
            "Note that the counts for these channels may be out of range, "
            "please check your normalization settings.".format(fov, abnormal_increase_chans)
        )
        norm_vals[abnormal_increase_mask] = 0.1

    decrease_mask = np.where(norm_vals > 1.1)[0]
    if len(decrease_mask) > 0:
        decrease_chans = np.array(channels)[decrease_mask]
        warnings.warn(
            "The following channel(s) are at greater than 110% sensitivity, which may indicate "
            "that your detector was overgained for this run for fov {}. "
            "Normalization capped at 0.9X for: {}.".format(fov, decrease_chans)
        )
        norm_vals[decrease_mask] = 1.1

    # check if any additional values are outside expected range
    extreme_mask = np.where(np.logical_or(norm_vals < extreme_vals[0], norm_vals > extreme_vals[1]))

    # remove channels that have been filtered by abnormal_increase_mask and decrease_mask
    extreme_mask = np.setdiff1d(extreme_mask, np.union1d(abnormal_increase_mask, decrease_mask))

    if len(extreme_mask) > 0:
        bad_channels = np.array(channels)[extreme_mask]
        warnings.warn(
            "The following additional channel(s) had an extreme normalization "
            "value for fov {}. Manual inspection for accuracy is "
            "recommended: {}".format(fov, bad_channels)
        )

    # correct images and save, ensure that no division by zero happens
    norm_vals = norm_vals.astype(img_data.dtype)
    norm_div = norm_vals.reshape((1, 1, 1, len(norm_vals)))
    normalized_images = np.ma.divide(img_data.values, norm_div).filled(0)

    for idx, chan in enumerate(channels):
        fname = os.path.join(output_fov_dir, chan + ".tiff")
        image_utils.save_image(fname, normalized_images[0, :, :, idx])

    # save logs
    log_df = pd.DataFrame({"channels": channels, "norm_vals": norm_vals})
    log_df.to_csv(os.path.join(output_fov_dir, "normalization_coefs.csv"), index=False)


def normalize_image_data(
    img_dir,
    norm_dir,
    pulse_height_dir,
    panel_info,
    img_sub_folder="",
    mass_obj_func="poly_2",
    extreme_vals=(0.4, 1.1),
    norm_func_path=os.path.join("..", "tuning_curves", "avg_norm_func_2600.json"),
    autogain=False,
    **kwargs,
):
    """Normalizes image data based on median pulse height from the run and a tuning curve.

    Args:
        img_dir (str): directory with the image data
        norm_dir (str): directory where the normalized images will be saved
        pulse_height_dir (str): directory containing per-fov pulse heights
        panel_info (pd.DataFrame): mapping between channels and masses
        img_sub_folder (str): subfolder containing images
        mass_obj_func (str): class of function to use for modeling MPH over time per mass
        extreme_vals (tuple): determines the range for norm vals which will raise a warning
        norm_func_path (str): file containing the saved weights for the normalization function
        autogain (bool): whether the run had autogain turned on or not
        **kwargs (dict): additional parameters to pass to `fit_mass_mph_curve` for outlier detection
    """
    # error checks
    if not os.path.exists(norm_func_path):
        raise ValueError(
            "No normalization function found. You will need to run "
            "section 3 of the 1_set_up_toffy.ipynb notebook to generate the "
            "necessary function before you can normalize your data"
        )

    # create normalization function for mapping MPH to counts
    norm_json = read_json_file(norm_func_path)

    # TODO: list_folders does not handle cases such as "fov0" correctly
    # need to add a fix in to alpineer to deal with this
    img_fovs = [f for f in os.listdir(img_dir) if "fov" in f]

    norm_weights, norm_name = norm_json["weights"], norm_json["name"]
    norm_func = create_prediction_function(norm_name, norm_weights)

    # combine pulse heights together into single df
    pulse_height_df = create_fitted_pulse_heights_file(
        pulse_height_dir=pulse_height_dir,
        panel_info=panel_info,
        norm_dir=norm_dir,
        mass_obj_func=mass_obj_func,
        autogain=autogain,
        kwargs=kwargs,
    )
    # add channel name to pulse_height_df
    renamed_panel = panel_info.rename({"Mass": "mass"}, axis=1)
    pulse_height_df = pulse_height_df.merge(renamed_panel, how="left", on=["mass"])
    pulse_height_df = pulse_height_df.sort_values("Target")

    # make sure FOVs used to construct tuning curve are same ones being normalized
    pulse_fovs = np.unique(pulse_height_df["fov"])

    # TODO: verify_same_elements needs to throw a ValueError in the special case
    # where one list is empty but the other isn't
    misc_utils.verify_same_elements(image_data_fovs=img_fovs, pulse_height_csv_files=pulse_fovs)

    # loop over each fov
    for fov in img_fovs:
        # compute per-mass normalization constant
        pulse_height_fov = pulse_height_df.loc[pulse_height_df["fov"] == fov, :]
        channels = pulse_height_fov["Target"].values
        norm_vals = norm_func(pulse_height_fov["pulse_height_fit"].values)

        # get images
        images = load_utils.load_imgs_from_tree(
            img_dir, fovs=[fov], channels=channels, img_sub_folder=img_sub_folder
        )

        # normalize and save
        normalize_fov(
            img_data=images,
            norm_vals=norm_vals,
            norm_dir=norm_dir,
            fov=fov,
            channels=channels,
            extreme_vals=extreme_vals,
        )


def check_detector_voltage(run_dir):
    """Check all FOVs in a run to determine whether the detector voltage stays constant.

    Args:
        run_dir(string): path to directory containing json files of all fovs in the run
    Return:
        raise error if changes in voltage were found between fovs
    """
    fovs = io_utils.remove_file_extensions(io_utils.list_files(run_dir, substrs=".bin"))
    changes_in_voltage = []

    # skip any damaged fovs
    empty_fovs = check_for_empty_files(run_dir)
    fovs = list(set(fovs).difference(empty_fovs))
    fovs = ns.natsorted(fovs)

    # check for voltage changes and add to list of dictionaries
    for i, fov in enumerate(fovs):
        fov_data = read_json_file(os.path.join(run_dir, fov + ".json"))

        # locate index storing the detector voltage
        for j in range(0, len(fov_data["hvDac"])):
            if fov_data["hvDac"][j]["name"] == "Detector":
                index = j
                break
        fov_voltage = fov_data["hvDac"][index]["currentSetPoint"]

        if i == 0:
            voltage_level = fov_voltage

        # detector voltage for current fov is different than previous
        elif fov_voltage != voltage_level:
            changes_in_voltage.append({fovs[i - 1]: voltage_level, fovs[i]: fov_voltage})
            voltage_level = fov_voltage

    err_str = ""
    for i, change in enumerate(changes_in_voltage):
        keys = list(change.keys())
        err_i = "Between {0} and {1} the voltage changed from {2} to {3}.".format(
            keys[0], keys[1], change[keys[0]], change[keys[1]]
        )
        err_str = err_str + "\n" + err_i

    # non-empty list of changes will raise an error
    if changes_in_voltage:
        raise ValueError("Changes in detector voltage were found during the run:\n" + err_str)


def plot_detector_voltage(
    run_folder: Union[pathlib.Path, str],
    mph_run_dir: Union[pathlib.Path, str],
    detector_volt_type: Literal["hvAdc", "hvDac"] = "hvAdc",
) -> None:
    """Extracts saved detector voltage values during fov acquisition and creates a txt file and png plot with
    detector voltage over fov progression.

    Args:
        run_folder (Union[pathlib.Path, str]):
            Path to where the run data (and json files) are stored.
        mph_run_dir (Union[pathlib.Path, str]):
            Path to where the run metrics are stored.
        detector_volt_type (Union["hvAdc", "hvDac"]):
            Either "hvAdc" (default) or "hvDac" voltage metric to plot.
    """
    # Verify that detector_volt_type is valid
    misc_utils.verify_in_list(
        specified_volt_type=detector_volt_type, valid_volt_types=["hvAdc", "hvDac"]
    )

    # Get all the JSON run files
    json_files = sorted(glob.glob(os.path.join(run_folder, "fov-*-scan-?.json")))

    # Map each FOV to their observed detector voltage
    detector_voltage_data = {"FOVs": [], "Detector Voltage": []}

    # Iterate over the list of files
    for json_file in json_files:
        # Load the JSON file
        run_json_data = read_json_file(json_file, encoding="utf-8")

        # Extract the FOV from the filename
        fov = os.path.basename(json_file).split("-")[1]

        # For hvAdc and hvDac, the key associated with the voltage is different
        voltage_key = "value" if detector_volt_type == "hvAdc" else "currentSetPoint"

        # Extract the Detector voltage
        # Assuming the "hvAdc" and/or "hvDac" field is always present and has the "Detector" entry
        detector_voltage = [
            item[voltage_key]
            for item in run_json_data[detector_volt_type]
            if item["name"] == "Detector"
        ][0]

        # Add the data to our dictionary
        detector_voltage_data["FOVs"].append(int(fov))
        detector_voltage_data["Detector Voltage"].append(detector_voltage)

    # Save data in a text log
    detector_gain_file = os.path.join(mph_run_dir, "detector_gain_values.csv")
    detector_gain_df = pd.DataFrame(detector_voltage_data)
    detector_gain_df.to_csv(detector_gain_file, index=False)

    # Create a new figure and axes
    fig, ax = plt.subplots()

    # Plot the data with color coding and without a line connecting the points
    for (
        fov,
        voltage,
    ) in zip(detector_voltage_data["FOVs"], detector_voltage_data["Detector Voltage"]):
        ax.plot(fov, voltage, marker="o", color="blue")

    # Set the labels for the axes
    ax.set_xlabel("FOVs")
    ax.set_ylabel("Detector Voltage (v)")

    # Show the plot
    plt.savefig(
        fname=os.path.join(mph_run_dir, "detector_gain_plot.png"),
        dpi="figure",
        format="png",
        metadata=None,
        bbox_inches=None,
        pad_inches=0.1,
        facecolor="auto",
        edgecolor="auto",
        backend=None,
    )



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from alpineer import io_utils, misc_utils
from mibi_bin_tools.bin_files import get_histograms_per_tof


def bin_array(arr, bin_factor):
    """Bin data for visualization according to a bin_factor.

    Args:
        arr (array_like): original data
        bin_factor (int): size of the bins for the histograms

    Returns:
        the binned data

    """
    arr_bin = np.cumsum(arr)
    arr_bin[bin_factor:] = arr_bin[bin_factor:] - arr_bin[:-bin_factor]
    arr_bin = arr_bin[bin_factor::bin_factor]

    return arr_bin


def compute_intensities(bin_file_dir, fov_list, mass, mass_start, mass_stop, bin_factor=100):
    """Compute the pulse height intensities for given FOVs.

    Args:
        bin_file_dir (str): path to the FOV bin files
        fov_list (list): which FOVs to include
        mass (float): mass for the panel
        mass_start (float): beginning of mass integration range
        mass_stop (float): end of mass integration range
        bin_factor (int): size of the bins for the histograms, default 100

    Returns:
        pd.DataFrame containing the MPH intensity data

    """
    panel = pd.DataFrame(
        [
            {
                "Mass": mass,
                "Target": None,
                "Start": mass_start,
                "Stop": mass_stop,
            }
        ]
    )

    out_df = []

    # retrieve and store data for each fov
    for fov in fov_list:
        _, intensities, pulse_counts = get_histograms_per_tof(bin_file_dir, fov, None, panel)

        int_bin = np.cumsum(intensities) / intensities.sum()
        median = (np.abs(int_bin - 0.5)).argmin()

        out_df.append(
            {
                "fov": fov,
                "all_intensities": intensities,
                "all_pulse_counts": pulse_counts,
                "median_intensity": median,
            }
        )
    final_df = pd.DataFrame(out_df)

    # create column of binned intensity values and output df to csv
    final_df["binned_intensities"] = final_df["all_intensities"].apply(
        lambda x: bin_array(x, bin_factor)
    )
    return final_df


def visualize_intensity_data(
    bin_file_dir,
    mass,
    mass_start,
    mass_stop,
    fov_list=None,
    bin_factor=100,
    x_cutoff=20000,
    normalize=True,
):
    """Create a histogram of the pulse height intensities for given FOVs.

    Args:
        bin_file_dir (str): path to the FOV bin files
        mass (float): mass for the panel
        mass_start (float): beginning of mass integration range
        mass_stop (float): end of mass integration range
        fov_list (list): which FOVs to include, if None will include all in data_dir
        bin_factor (int): size of the bins for the histograms, default 100
        x_cutoff (int): cutoff for x-axis
        normalize (bool): whether to normalize the histograms

    """
    # validate path
    io_utils.validate_paths(bin_file_dir)

    # compute for all FOVs in folder if list not specified, verify valid fov names
    all_fovs = io_utils.remove_file_extensions(io_utils.list_files(bin_file_dir, substrs=".bin"))
    if fov_list is None:
        fov_list = all_fovs
    else:
        misc_utils.verify_in_list(provided_fovs=fov_list, fovs_in_directory=all_fovs)

    # compute the pulse height intensities
    data = compute_intensities(bin_file_dir, fov_list, mass, mass_start, mass_stop, bin_factor)

    # plot each FOV
    plt.style.use("dark_background")
    for idx, row in data.iterrows():
        if normalize:
            plt.plot(
                np.arange(row["binned_intensities"].shape[0])[0 : x_cutoff // bin_factor]
                * bin_factor,
                row["binned_intensities"][0 : x_cutoff // bin_factor]
                / np.max(row["binned_intensities"]),
            )
        else:
            plt.plot(
                np.arange(row["binned_intensities"].shape[0])[0 : x_cutoff // bin_factor]
                * bin_factor,
                row["binned_intensities"][0 : x_cutoff // bin_factor],
            )

    plt.gca().set_xlabel("pulse height")
    plt.gca().set_ylabel("occurrence count")
    plt.gcf().set_size_inches(18.5, 10.5)
